(* seplog (c) AIST 2005-2015. R. Affeldt, N. Marti, et al. GNU GPLv3. *)
From mathcomp Require Import ssreflect ssrfun ssrbool eqtype ssrnat seq.
Require Import ssrZ ZArith_ext seq_ext uniq_tac machine_int multi_int omegaz.
Import MachineInt.
Require Import mapstos mips_seplog mips_contrib mips_tactics mont_mul_prg.
Import expr_m.
Import assert_m.

Local Open Scope machine_int_scope.
Local Open Scope eqmod_scope.
Local Open Scope zarith_ext_scope.
Local Open Scope heap_scope.
Local Open Scope mips_expr_scope.
Local Open Scope mips_assert_scope.
Local Open Scope mips_cmd_scope.
Local Open Scope mips_hoare_scope.
Local Open Scope uniq_scope.
Local Open Scope multi_int_scope.

(** printing `* %\ensuremath{\odot}% *)

(** Formal separation-logic triple that specifies the Montgomery
multiplication.  In the post-condition, observe that the %\ensuremath{k+1}th%#k+1
th# word of storage is provided by the register %\coqdocvar{C}%#C#. The
existence in memory of input-words and outputs-words is specified by
a separation logic formula. *)

Lemma mont_mul_triple (k alpha x y z m one ext int_ X_ Y_ M_ Z_ quot C t s_ : reg) :
  uniq(k, alpha, x, y, z, m, one, ext, int_, X_, Y_, M_, Z_, quot, C, t, s_, r0) ->
 forall nk valpha vx vy vm vz X Y M,
  u2Z (M `32_ 0) * u2Z valpha =m -1 {{ \B^1 }} ->
  size X = nk -> size Y = nk -> size M = nk ->
  u2Z vz + 4 * Z_of_nat nk < \B^1 ->
  \S_{ nk } X < \S_{ nk } M -> \S_{ nk } Y < \S_{ nk } M ->
{{ fun s h => [x]_s = vx /\ [y]_s = vy /\ [z]_s = vz /\ [m]_s = vm /\ 
  u2Z [k]_s = Z_of_nat nk /\ [alpha]_s = valpha /\
  (var_e x |--> X ** var_e y |--> Y ** var_e z |--> nseq nk zero32 ** var_e m |--> M) s h /\
  store.multi_null s }}
montgomery k alpha x y z m one ext int_ X_ Y_ M_ Z_ quot C t s_
{{ fun s h => exists Z, size Z = nk /\
  [x]_s = vx /\ [y]_s = vy /\ [z]_s = vz /\ [m]_s = vm /\
  u2Z [k]_s = Z_of_nat nk /\ [alpha]_s = valpha /\
  (var_e x |--> X ** var_e y |--> Y ** var_e z |--> Z ** var_e m |--> M) s h /\
  \B^nk * \S_{ nk.+1 } (Z ++ [C]_s :: nil) =m 
    \S_{ nk } X * \S_{ nk } Y {{ \S_{ nk } M }} /\
  \S_{ nk.+1 } (Z ++ [C]_s :: nil) < 2 * \S_{ nk } M /\ 
  u2Z [t]_s = u2Z vz + 4 * Z_of_nat nk.-1 }}.
Proof.
move=> Hset nk valpha vx vy vm vz X Y M Halpha Hx Hy Hm Hnz HX HY. 
rewrite /montgomery.

(** addiu one r0 one16; *)

NextAddiu.
move=> s h [r_x [r_y [r_z [r_m [r_k [r_alpha [Hmem Hmu]]]]]]].
rewrite /wp_addiu.
repeat Reg_upd; repeat (split; trivial).
by Assert_upd.
by rewrite store.multi_null_upd.

(** addiu C r0 zero16; *)

NextAddiu.
move=> s h [[r_x [r_y [r_z [r_m [r_k [r_alpha [Hmem Hmu]]]]]]] r_one].
rewrite /wp_addiu.
repeat Reg_upd; repeat (split; trivial).
by Assert_upd.
by rewrite store.multi_null_upd.

(** addiu ext r0 zero16; *)

NextAddiu.
move=> s h [[[r_x [r_y [r_z [r_m [r_k [rd_galpha [Hmem Hmu]]]]]]] r_one] r_C].
rewrite /wp_addiu.
repeat Reg_upd; repeat (split; trivial).
by Assert_upd.
by rewrite store.multi_null_upd.

(** while (bne ext k) *)

apply hoare_prop_m.hoare_while_invariant with (fun s h => exists Z next, size Z = nk /\
  [x]_s = vx /\ [y]_s = vy /\ [z]_s = vz /\ [m]_s = vm /\
  u2Z [k]_s = Z_of_nat nk /\ [alpha]_s = valpha /\
  (var_e x |--> X ** var_e y |--> Y ** var_e z |--> Z ** var_e m |--> M) s h /\
  u2Z [ext]_s = Z_of_nat next /\ u2Z [ext]_s <= u2Z [k]_s /\
  [one]_s = one32 /\ (next <> O -> u2Z [t]_s = u2Z vz + 4 * Z_of_nat nk.-1) /\
  (exists K, \B^next * \S_{ nk.+1 } (Z ++ [C]_s :: nil) = 
               \S_{next} X * \S_{nk} Y + K * \S_{nk} M /\
    \S_{next} X * \S_{nk} Y + K * \S_{nk} M < 2 * \S_{nk} M * \B^next)).

move=> s h [[[[r_x [r_y [r_z [r_m [r_k [r_alpha [Hmem Hmu]]]]]]] r_one] r_C] r_ext].

exists (nseq nk zero32), O; repeat (split; trivial).
by apply size_nseq.
by rewrite r_ext sext_Z2u // addi0 store.get_r0 Z2uK.
rewrite r_ext sext_Z2u // addi0 store.get_r0 Z2uK //; by apply min_u2Z.
apply u2Z_inj; by rewrite r_one sext_Z2u // store.get_r0 add0i Z2uK.
by case.
exists 0.
have -> : nseq nk zero32 ++ [C ]_ s :: nil = nseq nk.+1 zero32.
  suff -> : [ C ]_s = zero32 by rewrite -nseqS.
  apply u2Z_inj; by rewrite r_C sext_Z2u // addi0 store.get_r0.
rewrite lSum_nseq_0.
split; first by done.
(* about X * Y < M *)
rewrite Zmult_1_r !Zmult_0_l Zplus_0_l.
apply Zmult_lt_0_compat; first by done.
apply Zle_lt_trans with (\S_{ nk } X); [by apply min_lSum | assumption].

move=> s h [[Z [next [HZ [r_x [r_y [r_z [r_m [r_k [r_alpha [Hmem [Hext [Hextk' [Hone [Ht Hinv]]]]]]]]]]]]]] Hextk].
rewrite /= Hext r_k in Hextk; move/negPn/ZeqP/Z_of_nat_inj in Hextk; subst next.

exists Z; do 8 (split; trivial).
case: Hinv => K [Hinv1 Hinv2].
split; first by exists K.
split.
(* about X * Y < M *)
- apply Zmult_gt_0_lt_reg_r with \B^nk.
  by apply Zlt_gt, Zbeta_0.
  rewrite Zmult_comm Hinv1; exact Hinv2.
- apply Ht; by destruct nk.

(** lwxs X ext x; *)

apply hoare_lwxs_back_alt'' with (fun s h => exists Z next, size Z = nk /\
  [x]_s = vx /\ [y]_s = vy /\ [z]_s = vz /\ [m]_s = vm /\
  u2Z [k]_s = Z_of_nat nk /\ [alpha]_s = valpha /\
  (var_e x |--> X ** var_e y |--> Y ** var_e z |--> Z ** var_e m |--> M) s h /\
  u2Z [ext]_s = Z_of_nat next /\ u2Z [ext]_s < u2Z [k]_s /\ [one]_s = one32 /\
  (exists K, \B^next * \S_{ nk.+1 } (Z ++ [C]_s :: nil) = \S_{ next } X * \S_{ nk } Y + K * \S_{ nk } M /\
    \S_{ next } X * \S_{ nk } Y + K * \S_{ nk } M < 2 * \S_{ nk } M * \B^next) /\
  (next < nk)%nat /\ [X_]_s = X `32_ next).

move=> s h [[Z [next [HZ [r_x [r_y [r_z [r_m [r_k [r_alpha [Hmem [r_ext [Hextk' [Hone [Ht Hinv]]]]]]]]]]]]]] Hextk]. 
rewrite /= in Hextk; move/ZeqP in Hextk.
have {Hextk} Hextk : (next < nk)%nat.
  apply/ltP/Nat2Z.inj_lt; rewrite -r_ext -r_k; by apply Zle_neq_lt.
exists (X `32_ next); split.
- Decompose_32 X next X1 X2 HlenX1 HX'; last by rewrite Hx.
  rewrite HX' (decompose_equiv _ _ _ _ _ HlenX1) !assert_m.conAE assert_m.conCE !assert_m.conAE in Hmem.
  move: Hmem; apply monotony => // h'.
  apply mapsto_ext => //.
  by rewrite /= shl_Z2u r_ext inj_mult Zmult_comm.
- rewrite /update_store_lwxs.
  exists Z, next; repeat Reg_upd; repeat (split; trivial).
  by Assert_upd.
  rewrite r_ext r_k; by apply/inj_lt/ltP.

(** lw Y zero16 y; *)

apply hoare_lw_back_alt'' with (fun s h => exists Z next, size Z = nk /\
  [x]_s = vx /\ [y]_s = vy /\ [z]_s = vz /\ [m]_s = vm /\
  u2Z [k]_s = Z_of_nat nk /\ [alpha]_s = valpha /\
  (var_e x |--> X ** var_e y |--> Y ** var_e z |--> Z ** var_e m |--> M) s h /\
  u2Z [ext]_s = Z_of_nat next /\ u2Z [ext]_s < u2Z [k]_s /\ [one]_s = one32 /\ 
  (exists K, \B^next * \S_{nk.+1} (Z ++ [C]_s :: nil) = \S_{ next } X * \S_{ nk } Y + K * \S_{ nk } M /\
    \S_{ next } X * \S_{ nk } Y + K * \S_{ nk } M < 2 * \S_{ nk } M * \B^next) /\
  (next < nk)%nat /\ [X_]_s = X `32_ next /\ [Y_]_s = Y `32_ 0).

move=> s h [Z [next [HZ [r_x [r_y [r_z [r_m [r_k [r_alpha [mem [r_ext [Hextk' [Hone [Hinv [Hextk r_X_]]]]]]]]]]]]]]].

destruct Y as [| hdy tly]; first by destruct nk.
exists hdy; split.
rewrite /= !assert_m.conAE assert_m.conCE assert_m.conAE in mem.
move: mem; apply monotony => ht //.
apply mapsto_ext => //; by rewrite /= sext_0 addi0.
rewrite /update_store_lw.
exists Z, next; repeat Reg_upd; repeat (split; trivial).
by Assert_upd.

(** lw Z_ zero16 z; *)

apply hoare_lw_back_alt'' with (fun s h => exists Z next, size Z = nk /\
  [x]_s = vx /\ [y]_s = vy /\ [z]_s = vz /\ [m]_s = vm /\
  u2Z [k]_s = Z_of_nat nk /\ [alpha]_s = valpha /\
  (var_e x |--> X ** var_e y |--> Y ** var_e z |--> Z ** var_e m |--> M) s h /\
  u2Z [ext]_s = Z_of_nat next /\ u2Z [ext]_s < u2Z [k]_s /\ [one]_s = one32 /\
  (exists K, \B^next * \S_{ nk.+1 } (Z ++ [C]_s :: nil) = \S_{ next } X * \S_{ nk } Y + K * \S_{ nk } M /\
    \S_{ next } X * \S_{ nk } Y + K * \S_{ nk } M < 2 * \S_{ nk } M * \B^next) /\
  (next < nk)%nat /\ [X_]_s = X `32_ next /\ 
  [Y_]_s = Y `32_ 0 /\ [Z_]_s = Z `32_ 0).

move=> s h [Z [next [HZ [r_x [r_y [r_z [r_m [r_k [r_alpha [mem [r_ext [Hextk' [Hone [Hinv [Hextk [r_X_ r_Y_]]]]]]]]]]]]]]]].

destruct Z as [| hdz tlz]; first by destruct nk.
exists hdz; split.
rewrite /= in mem.
do 2 rewrite !assert_m.conAE assert_m.conCE in mem; rewrite !assert_m.conAE in mem.
move: mem; apply monotony => ht //.
apply mapsto_ext => //; by rewrite /= sext_0 addi0.
rewrite /update_store_lw.
exists (hdz :: tlz), next; repeat Reg_upd; repeat (split; trivial).
by Assert_upd.

(** multu X Y; *)

apply hoare_multu with (fun s h => exists Z next, size Z = nk /\
  [x]_s = vx /\ [y]_s = vy /\ [z]_s = vz /\ [m]_s = vm /\
  u2Z [k]_s = Z_of_nat nk /\ [alpha]_s = valpha /\
  (var_e x |--> X ** var_e y |--> Y ** var_e z |--> Z ** var_e m |--> M) s h /\
  u2Z [ext]_s = Z_of_nat next /\ u2Z [ext]_s < u2Z [k]_s /\ [one]_s = one32 /\ 
  (exists K, \B^next * \S_{ nk.+1 } (Z ++ [C]_s :: nil) = \S_{ next } X * \S_{ nk } Y + K * \S_{ nk } M /\
    \S_{ next } X * \S_{ nk } Y + K * \S_{ nk } M < 2 * \S_{ nk } M * \B^next) /\
  (next < nk)%nat /\ [X_]_s = X `32_ next /\ [Y_]_s = Y `32_ 0 /\ [Z_]_s = Z `32_ 0 /\
  store.utoZ s <= (\B^1 - 1) * (\B^1 - 1) /\ store.utoZ s = u2Z (X `32_ next) * u2Z (Y `32_ 0)).

move=> s h [Z [next [HZ [r_x [r_y [r_z [r_m [r_k [r_alpha [Hmem [r_ext [Hextk' [Hone [Hinv [Hextk [r_X_ [r_Y_ r_Z_]]]]]]]]]]]]]]]]].

exists Z, next; repeat Reg_upd; repeat (split; trivial).
by Assert_upd.

by rewrite store.utoZ_multu; apply max_u2Z_umul.
by rewrite store.utoZ_multu (@u2Z_umul 32) r_X_ r_Y_.

(** lw M zero16 m; *)

apply hoare_lw_back_alt'' with (fun s h => exists Z next, size Z = nk /\
  [x]_s = vx /\ [y]_s = vy /\ [z]_s = vz /\ [m]_s = vm /\
  u2Z [k]_s = Z_of_nat nk /\ [alpha]_s = valpha /\
  (var_e x |--> X ** var_e y |--> Y ** var_e z |--> Z ** var_e m |--> M) s h /\
  u2Z [ext]_s = Z_of_nat next /\ u2Z [ext]_s < u2Z [k]_s /\ [one]_s = one32 /\ 
  (exists K, \B^next * \S_{ nk.+1 } (Z ++ [C]_s :: nil) = \S_{ next } X * \S_{ nk } Y + K * \S_{ nk } M /\
    \S_{ next } X * \S_{ nk } Y + K * \S_{ nk } M < 2 * \S_{ nk } M * \B^next) /\
  (next < nk)%nat /\ [X_]_s = X `32_ next /\ [Y_]_s = Y `32_ 0 /\
  [Z_]_s = Z `32_ 0 /\ store.utoZ s <= (\B^1 - 1) * (\B^1 - 1) /\
  store.utoZ s = u2Z (X `32_ next) * u2Z (Y `32_ 0) /\ [M_]_s = M `32_ 0).

move=> s h [Z [next [HZ [r_x [r_y [r_z [r_m [r_k [r_alpha [mem [r_ext [Hextk' [Hone [Hinv [Hextk [r_X_ [r_Y_ [r_Z_ [Hm1 Hm2]]]]]]]]]]]]]]]]]]].

destruct M as [| hdm tlm]; first by destruct nk.
exists hdm; split.
rewrite /= in mem.
do 3 rewrite assert_m.conCE !assert_m.conAE in mem.
move: mem; apply monotony => h' //.
apply mapsto_ext => //=; by rewrite sext_0 addi0.
exists Z, next; repeat Reg_upd; repeat (split; trivial).
by Assert_upd.

(** maddu Z_ one; *)

apply hoare_maddu with (fun s h => exists Z next, size Z = nk /\
  [x]_s = vx /\ [y]_s = vy /\ [z]_s = vz /\ [m]_s = vm /\
  u2Z [k]_s = Z_of_nat nk /\ [alpha]_s = valpha /\
  (var_e x |--> X ** var_e y |--> Y ** var_e z |--> Z ** var_e m |--> M) s h /\
  u2Z [ext]_s = Z_of_nat next /\ u2Z [ext]_s < u2Z [k]_s /\ [one]_s = one32 /\ 
  (exists K, \B^next * \S_{ nk.+1 } (Z ++ [C]_s :: nil) = \S_{ next } X * \S_{ nk } Y + K * \S_{ nk } M /\
    \S_{ next } X * \S_{ nk } Y + K * \S_{ nk } M < 2 * \S_{ nk } M * \B^next) /\
  (next < nk)%nat /\ [X_]_s = X `32_ next /\ [Y_]_s = Y `32_ 0 /\ [Z_]_s = Z `32_ 0 /\
  store.utoZ s <= \B^1 * (\B^1 - 1) /\
  store.utoZ s = u2Z (X `32_ next) * u2Z (Y `32_ 0) + u2Z (Z `32_ 0) /\ [M_]_s = M `32_ 0).

move=> s h [Z [next [HZ [r_x [r_y [r_z [r_m [r_k [r_alpha [Hmem [r_ext [Hextk' [Hone [Hinv [Hextk [r_X_ [r_Y_ [r_Z_ [Hm1 [Hm2 r_M_]]]]]]]]]]]]]]]]]]]].
exists Z, next.
have Htmp : store.utoZ s < \B^2 * (2 ^^ store.acx_size - 1).
  by apply Zle_lt_trans with ((\B^1 - 1) * (\B^1 - 1)).
repeat Reg_upd.
do 8 (split; trivial).
by Assert_upd.
do 8 (split; trivial).
split.
- rewrite store.utoZ_maddu // Hone umul_1 (@u2Z_zext 32) r_Z_.
  apply Zle_trans with ((\B^1 - 1) + (\B^1 - 1) * (\B^1 - 1)); last by done.
  apply Zplus_le_compat.
  by apply Zle_minus_1; rewrite /Zbeta; apply max_u2Z.
  rewrite Hm2 -u2Z_umul; by apply max_u2Z_umul.
- by rewrite store.utoZ_maddu // Hone umul_1 (@u2Z_zext 32) r_Z_ Hm2 Zplus_comm.

(** mflo t; *)

apply hoare_mflo with (fun s h => exists Z next, size Z = nk /\
  [x]_s = vx /\ [y]_s = vy /\ [z]_s = vz /\ [m]_s = vm /\
  u2Z [k]_s = Z_of_nat nk /\ [alpha]_s = valpha /\
  (var_e x |--> X ** var_e y |--> Y ** var_e z |--> Z ** var_e m |--> M) s h /\
  u2Z [ext]_s = Z_of_nat next /\ u2Z [ext]_s < u2Z [k]_s /\ [one]_s = one32 /\ 
  (exists K, \B^next * \S_{ nk.+1 } (Z ++ [C]_s :: nil) = \S_{ next } X * \S_{ nk } Y + K * \S_{ nk } M /\
    \S_{ next } X * \S_{ nk } Y + K * \S_{ nk } M < 2 * \S_{ nk } M * \B^next) /\
  (next < nk)%nat /\ [X_]_s = X `32_ next /\ [Y_]_s = Y `32_ 0 /\ [Z_]_s = Z `32_ 0 /\
  store.utoZ s <= \B^1 * (\B^1 - 1) /\
  store.utoZ s = u2Z (X `32_ next) * u2Z (Y `32_ 0) + u2Z (Z `32_ 0) /\ [M_]_s = M `32_ 0 /\
  [t]_s = ((X `32_ next `* Y `32_ 0) `+ zext 32 (Z `32_ 0)) `% 32).

move=> s h [Z [next [HZ [r_x [r_y [r_z [r_m [r_k [r_alpha [Hmem [r_ext [Hextk' [Hone [Hinv [Hextk [r_X_ [r_Y_ [r_Z_ [Hm1 [Hm2 r_M_]]]]]]]]]]]]]]]]]]]].

exists Z, next; repeat Reg_upd; repeat (split; trivial).
by Assert_upd.
apply store.lo_remainder; first by rewrite -addnA leq_addl.
rewrite Hm2 u2Z_add // u2Z_umul u2Z_zext // ZpowerD -Zbeta1_Zpower2 -Zbeta_is_exp /=.
apply Zle_lt_trans with ((\B^1 - 1) * (\B^1 - 1) + (\B^1 - 1)); last by done.
apply Zplus_le_compat.
rewrite -u2Z_umul; by apply max_u2Z_umul.
apply Zle_minus_1; by apply max_u2Z.

(** mfhi s; *)

apply hoare_mfhi with (fun s h => exists Z next, size Z = nk /\
  [x]_s = vx /\ [y]_s = vy /\ [z]_s = vz /\ [m]_s = vm /\
  u2Z [k]_s = Z_of_nat nk /\ [alpha]_s = valpha /\
  (var_e x |--> X ** var_e y |--> Y ** var_e z |--> Z ** var_e m |--> M) s h /\
  u2Z [ext]_s = Z_of_nat next /\ u2Z [ext]_s < u2Z [k]_s /\ [one]_s = one32 /\ 
  (exists K, \B^next * \S_{ nk.+1 } (Z ++ [C]_s :: nil) = \S_{ next } X * \S_{ nk } Y + K * \S_{ nk } M /\
    \S_{ next } X * \S_{ nk } Y + K * \S_{ nk } M < 2 * \S_{ nk } M * \B^next) /\
  (next < nk)%nat /\ [X_]_s = X `32_ next /\ [Y_]_s = Y `32_ 0 /\
  [Z_]_s = Z `32_ 0 /\ store.utoZ s <= \B^1 * (\B^1 - 1) /\
  store.utoZ s = u2Z (X `32_ next) * u2Z (Y `32_ 0) + u2Z (Z `32_ 0) /\
  [M_]_s = M `32_ 0 /\
  [t]_s = ((X `32_ next `* Y `32_ 0) `+ zext 32 (Z `32_ 0)) `% 32 /\
  u2Z ([s_]_s `|| [t]_s) = u2Z (X `32_ next) * u2Z (Y `32_ 0) + u2Z (Z `32_ 0)).

move=> s h [Z [next [HZ [r_x [r_y [r_z [r_m [r_k [r_alpha [Hmem [r_ext [Hextk' [Hone [Hinv [Hextk [r_X_ [r_Y_ [r_Z_ [Hm1 [Hm2 [r_M_ r_t]]]]]]]]]]]]]]]]]]]]].
exists Z, next; repeat Reg_upd; repeat (split; trivial).
by Assert_upd.
have H : store.lo s = [t]_s.
  rewrite r_t; apply store.lo_remainder.
  + by rewrite -addnA leq_addl.
  + rewrite Hm2 -u2Z_umul u2Z_add.
    * by rewrite u2Z_zext.
    * rewrite u2Z_zext ZpowerD -Zbeta1_Zpower2 -Zbeta_is_exp /=.
      apply Zle_lt_trans with ((\B^1 - 1) * (\B^1 - 1) + (\B^1 - 1)); last by done.
      apply Zplus_le_compat.
      - by apply (@max_u2Z_umul 32).
      - apply Zle_minus_1; by apply max_u2Z.
rewrite store.utoZ_def store.utoZ_acx_beta2 in Hm2; last by apply Zle_lt_trans with (\B^1 * (\B^1 - 1)).
rewrite Z2uK in Hm2; last by split; [apply Zle_refl | apply Zpower_0].
rewrite -H u2Z_concat -Zbeta1_Zpower2 -Hm2; ring.

(** multu t alpha; *)

apply hoare_multu with (fun s h => exists Z next, size Z = nk /\
  [x]_s = vx /\ [y]_s = vy /\ [z]_s = vz /\ [m]_s = vm /\
  u2Z [k]_s = Z_of_nat nk /\ [alpha]_s = valpha /\
  (var_e x |--> X ** var_e y |--> Y ** var_e z |--> Z ** var_e m |--> M) s h /\
  u2Z [ext]_s = Z_of_nat next /\ u2Z [ext]_s < u2Z [k]_s /\ [one]_s = one32 /\ 
  (exists K, \B^next * \S_{ nk.+1 } (Z ++ [C]_s :: nil) = \S_{ next } X * \S_{ nk } Y + K * \S_{ nk } M /\
    \S_{ next } X * \S_{ nk } Y + K * \S_{ nk } M < 2 * \S_{ nk } M * \B^next) /\
  (next < nk)%nat /\ [X_]_s = X `32_ next /\ [Y_]_s = Y `32_ 0 /\
  [Z_]_s = Z `32_ 0 /\ store.utoZ s <= (\B^1 - 1) * (\B^1 - 1) /\ [M_]_s = M `32_ 0 /\
  [t]_s = ((X `32_ next `* Y `32_ 0) `+ zext 32 (Z `32_ 0)) `% 32 /\
  u2Z ([s_]_s `|| [t]_s) = u2Z (X `32_ next) * u2Z (Y `32_ 0) + u2Z (Z `32_ 0) /\
  store.utoZ s = u2Z [t]_s * u2Z [alpha]_s).

move=> s h [Z [next [HZ [r_x [r_y [r_z [r_m [r_k [r_alpha [Hmem [r_ext [Hextk' [Hone [Hinv [Hextk [r_X_ [r_Y_ [r_Z_ [Hm1 [Hm2 [r_M_ [r_t Hconcat]]]]]]]]]]]]]]]]]]]]]].
exists Z, next; repeat Reg_upd.
do 7 (split; trivial).
split.
by Assert_upd.
do 8 (split; trivial).
split.
rewrite store.utoZ_multu; by apply max_u2Z_umul. 
do 3 (split; trivial).
by rewrite store.utoZ_multu -u2Z_umul.

(** addiu int_ r0 one16; *)

apply hoare_addiu with (fun s h => exists Z next, size Z = nk /\
  [x]_s = vx /\ [y]_s = vy /\ [z]_s = vz /\ [m]_s = vm /\
  u2Z [k]_s = Z_of_nat nk /\ [alpha]_s = valpha /\
  (var_e x |--> X ** var_e y |--> Y ** var_e z |--> Z ** var_e m |--> M) s h /\
  u2Z [ext]_s = Z_of_nat next /\ u2Z [ext]_s < u2Z [k]_s /\ [one]_s = one32 /\ 
  (exists K, \B^next * \S_{ nk.+1 } (Z ++ [C]_s :: nil) = \S_{ next } X * \S_{ nk } Y + K * \S_{ nk } M /\
    \S_{ next } X * \S_{ nk } Y + K * \S_{ nk } M < 2 * \S_{ nk } M * \B^next) /\
  (next < nk)%nat /\ [X_]_s = X `32_ next /\ [Y_]_s = Y `32_ 0 /\
  [Z_]_s = Z `32_ 0 /\ store.utoZ s <= (\B^1 - 1) * (\B^1 - 1) /\ [M_]_s = M `32_ 0 /\
  [t]_s = ((X `32_ next `* Y `32_ 0) `+ zext 32 (Z `32_ 0)) `% 32 /\
  u2Z ([s_]_s `|| [t]_s) = u2Z (X `32_ next) * u2Z (Y `32_ 0) + u2Z (Z `32_ 0) /\
  store.utoZ s = u2Z [t]_s * u2Z [alpha]_s /\ [int_]_s = one32).

move=> s h [Z [next [HZ [r_x [r_y [r_z [r_m [r_k [r_alpha [Hmem [r_ext [Hextk' [Hone [Hinv [Hextk [ r_X_ [r_Y_ [r_Z_ [Hm1 [r_M_ [r_t [Hconcat Hm2]]]]]]]]]]]]]]]]]]]]]].

exists Z, next; repeat Reg_upd; repeat (split; trivial).
by Assert_upd.
by rewrite add0i sext_Z2u.

(** mflo quot; *)

apply hoare_mflo with (fun s h => exists Z next, size Z = nk /\
  [x]_s = vx /\ [y]_s = vy /\ [z]_s = vz /\ [m]_s = vm /\
  u2Z [k]_s = Z_of_nat nk /\ [alpha]_s = valpha /\
  (var_e x |--> X ** var_e y |--> Y ** var_e z |--> Z ** var_e m |--> M) s h /\
  u2Z [ext]_s = Z_of_nat next /\ u2Z [ext]_s < u2Z [k]_s /\ [one]_s = one32 /\ 
  (exists K,
    \B^next * \S_{ nk.+1 } (Z ++ [C]_s :: nil) = \S_{ next } X * \S_{ nk } Y + K * \S_{ nk } M /\
    \S_{next} X * \S_{nk} Y + K * \S_{nk} M < 2 * \S_{ nk } M * \B^next) /\
  (next < nk)%nat /\ [X_]_s = X `32_ next /\
  [Y_]_s = Y `32_ 0 /\ [Z_]_s = Z `32_ 0 /\
  store.utoZ s <= (\B^1 - 1) * (\B^1 - 1) /\ [M_]_s = M `32_ 0 /\
  [t]_s = ((X `32_ next `* Y `32_ 0) `+ zext 32 (Z `32_ 0)) `% 32 /\
  u2Z ([s_]_s `|| [t]_s) = u2Z (X `32_ next) * u2Z (Y `32_ 0) + u2Z (Z `32_ 0) /\
  store.utoZ s = u2Z [t]_s * u2Z [alpha]_s /\ [int_]_s = one32 /\
  [quot]_s = (((X `32_ next `* Y `32_ 0) `+ (zext 32 (Z `32_ 0)) `% 32) `* [alpha]_s) `% 32).

move=> s h [Z [next [HZ [r_x [r_y [r_z [r_m [r_k [r_alpha [Hmem [r_ext [Hextk' [Hone [Hinv [Hextk [r_X_ [r_Y_ [ r_Z_ [Hm1 [r_M_ [r_t [Hconcat [Hm2 r_int_]]]]]]]]]]]]]]]]]]]]]]].
exists Z, next; repeat Reg_upd; repeat (split; trivial).
by Assert_upd.
apply store.lo_remainder.
- by rewrite -addnA leq_addl.
- by rewrite Hm2 r_t -u2Z_umul.

(** mthi s; *)

apply hoare_mthi with (fun s h => exists Z next, size Z = nk /\
  [x]_s = vx /\ [y]_s = vy /\ [z]_s = vz /\ [m]_s = vm /\
  u2Z [k]_s = Z_of_nat nk /\ [alpha]_s = valpha /\
  (var_e x |--> X ** var_e y |--> Y ** var_e z |--> Z ** var_e m |--> M) s h /\
  u2Z [ext]_s = Z_of_nat next /\ u2Z [ext]_s < u2Z [k]_s /\ [one]_s = one32 /\ 
  (exists K, \B^next * \S_{ nk.+1 } (Z ++ [C]_s :: nil) = \S_{ next } X * \S_{ nk } Y + K * \S_{ nk } M /\
    \S_{ next } X * \S_{ nk } Y + K * \S_{ nk } M < 2 * \S_{ nk } M * \B^next) /\
  (next < nk)%nat /\ [X_]_s = X `32_ next /\ [Y_]_s = Y `32_ 0 /\
  [Z_]_s = Z `32_ 0 /\ store.utoZ s < \B^2 /\ [M_]_s = M `32_ 0 /\
  [t]_s = ((X `32_ next `* Y `32_ 0) `+ zext 32 (Z `32_ 0)) `% 32 /\
  u2Z ([s_]_s `|| [t]_s) = u2Z (X `32_ next) * u2Z (Y `32_ 0) + u2Z (Z `32_ 0) /\
  [int_]_s = one32 /\
  [quot]_s = ((((X `32_ next `* Y `32_ 0) `+ zext 32 (Z `32_ 0)) `% 32) `* [alpha]_s) `% 32 /\ 
  store.hi s = [s_]_s).

move=> s h [Z [next [HZ [r_x [r_y [r_z [r_m [r_k [r_alpha [Hmem [r_ext [Hextk' [Hone [Hinv [Hextk [r_X_ [r_Y_ [ r_Z_ [Hm1 [r_M_ [r_t [Hconcat [Hm2 [r_int_ r_quot]]]]]]]]]]]]]]]]]]]]]]]].

exists Z, next; repeat Reg_upd; repeat (split; trivial).
by Assert_upd.
rewrite store.utoZ_def store.hi_mthi_op store.acx_mthi_op store.lo_mthi_op store.utoZ_acx_beta2; last first.
  rewrite Hm2.
  apply Zle_lt_trans with ((\B^1 - 1) * (\B^1 - 1)); last by done.
  rewrite -u2Z_umul; by apply max_u2Z_umul.
rewrite Z2uK //= -Zplus_0_r_reverse.
apply Zle_lt_trans with ((\B^1 - 1) + (\B^1 - 1) * \B^1); last by done.
apply Zplus_le_compat.
apply Zle_minus_1; by apply max_u2Z.
apply Zmult_le_compat_r.
apply Zle_minus_1; by apply max_u2Z.
by apply Zbeta_0'.
by apply store.hi_mthi_op.

(** mtlo t; *)

apply hoare_mtlo with  (fun s h => exists Z next, size Z = nk /\
  [x]_s = vx /\ [y]_s = vy /\ [z]_s = vz /\ [m]_s = vm /\
  u2Z [k]_s = Z_of_nat nk /\ [alpha]_s = valpha /\
  (var_e x |--> X ** var_e y |--> Y ** var_e z |--> Z ** var_e m |--> M) s h /\
  u2Z [ext]_s = Z_of_nat next /\ u2Z [ext]_s < u2Z [k]_s /\
  [one]_s = one32 /\ (exists K, \B^next * \S_{ nk.+1 } (Z ++ [C]_s :: nil) = \S_{ next } X * \S_{ nk } Y + K * \S_{ nk } M /\
    \S_{ next } X * \S_{ nk } Y + K * \S_{ nk } M < 2 * \S_{ nk } M * \B^next) /\
  (next < nk)%nat /\ [X_]_s = X `32_ next /\ [Y_]_s = Y `32_ 0 /\
  [Z_]_s = Z `32_ 0 /\ store.utoZ s < \B^2 /\ [M_]_s = M `32_ 0 /\
  [t]_s = ((X `32_ next `* Y `32_ 0) `+ zext 32 (Z `32_ 0)) `% 32 /\
  u2Z ([s_]_s `|| [t]_s) = u2Z (X `32_ next) * u2Z (Y `32_ 0) + u2Z (Z `32_ 0) /\
  [int_]_s = one32 /\
  [quot]_s = ((((X `32_ next `* Y `32_ 0) `+ zext 32 (Z `32_ 0)) `% 32) `* [alpha]_s) `% 32 /\
  store.hi s = [s_]_s /\ store.lo s = [t]_s).

move=> s h [Z [next [HZ [r_x [r_y [r_z [r_m [r_k [r_alpha [Hmem [r_ext [Hextk' [Hone [Hinv [Hextk [r_X_ [r_Y_ [r_Z_ [Hm1 [r_M_ [r_t [Hconcat [r_int_ [r_quot Hm2]]]]]]]]]]]]]]]]]]]]]]]].

exists Z, next; repeat Reg_upd; repeat (split; trivial).
by Assert_upd.
rewrite store.utoZ_def store.acx_mtlo_op store.lo_mtlo_op store.hi_mtlo_op store.utoZ_acx_beta2 //
  Z2uK //= -Zplus_0_r_reverse.
apply Zle_lt_trans with ((\B^1 - 1) + (\B^1 - 1) * \B^1); last by done.
apply Zplus_le_compat.
- apply Zle_minus_1; by apply max_u2Z.
- apply Zmult_le_compat_r; last by apply Zbeta_0'.
  apply Zle_minus_1; by apply max_u2Z.
rewrite -Hm2.
by apply store.hi_mtlo_op.
by apply store.lo_mtlo_op.

(** maddu quot M; *)

apply hoare_maddu with (fun s h => exists Z next, size Z = nk /\
  [x]_s = vx /\ [y]_s = vy /\ [z]_s = vz /\ [m]_s = vm /\
  u2Z [k]_s = Z_of_nat nk /\ [alpha]_s = valpha /\
  (var_e x |--> X ** var_e y |--> Y ** var_e z |--> Z ** var_e m |--> M) s h /\
  u2Z [ext]_s = Z_of_nat next /\ u2Z [ext]_s < u2Z [k]_s /\ [one]_s = one32 /\
  (exists K, \B^next * \S_{nk.+1} (Z ++ [C]_s :: nil) = \S_{ next } X * \S_{ nk } Y + K * \S_{ nk } M /\
    \S_{ next } X * \S_{ nk } Y + K * \S_{ nk } M < 2 * \S_{ nk } M * \B^next) /\
  (next < nk)%nat /\ [X_]_s = X `32_ next /\
  [Y_]_s = Y `32_ 0 /\ [Z_]_s = Z `32_ 0 /\
  [M_]_s = M `32_ 0 /\ [int_]_s = one32 /\
  [quot]_s = ((((X `32_ next `* Y `32_ 0) `+ zext 32 (Z `32_ 0)) `% 32) `* [alpha]_s) `% 32 /\
  store.utoZ s = u2Z [quot]_s * u2Z (M `32_ 0) + 
  u2Z (X `32_ next) * u2Z (Y `32_ 0) + u2Z (Z `32_ 0) /\
  store.lo s = zero32).

move=> s h [Z [next [HZ [r_x [r_y [r_z [r_m [r_k [r_alpha [Hmem [r_ext [Hextk' [Hone [Hinv [Hextk [r_X_ [r_Y_ [r_Z_ [Hm1 [r_M_ [r_t [Hconcat [r_int_ [r_quot [Hm2 Hm3]]]]]]]]]]]]]]]]]]]]]]]]].

exists Z, next; repeat Reg_upd; repeat (split; trivial).
by Assert_upd.
rewrite store.utoZ_maddu.
  rewrite r_M_ -u2Z_umul store.utoZ_def store.utoZ_acx_beta2 // Z2uK.
    by rewrite Hm2 Hm3 (Zplus_comm (u2Z [t]_s)) -u2Z_concat Hconcat Zmult_0_l -Zplus_0_r_reverse Zplus_assoc.
    by split; [apply Zle_refl | apply Zpower_0].
apply Zlt_le_trans with (\B^2 * 1); first by done.
apply Zmult_le_compat_l.
  apply Zle_trans with (2 ^^ 8 - 1); first by done.
  apply Zminus_le_compat.
  by apply/ZleP; rewrite Zpower_2_le store.acx_size_min.
  by apply Zbeta_0'.
rewrite r_quot; apply montgomery_lemma => //.
by rewrite r_M_ r_alpha.
rewrite store.utoZ_def store.utoZ_acx_beta2 // Z2uK.
- rewrite Zmult_0_l -Zplus_0_r_reverse Zplus_comm -u2Z_concat Hm2 Hm3 Hconcat -u2Z_umul u2Z_add.
  by rewrite (u2Z_zext 32) // u2Z_zext.
  apply Zle_lt_trans with ((\B^1 - 1) * (\B^1 - 1) + (\B^1 - 1)); last by done.
  apply Zplus_le_compat.
  + by apply (@max_u2Z_umul 32).
  + rewrite (u2Z_zext 32) //; apply Zle_minus_1; by apply max_u2Z.
- split; by [apply Zle_refl | apply Zpower_0].

(** mflhxu Z_; *)

apply hoare_mflhxu with (fun s h => exists Z next, size Z = nk /\
  [x]_s = vx /\ [y]_s = vy /\ [z]_s = vz /\ [m]_s = vm /\
  [one]_s = one32 /\ u2Z [k]_s = Z_of_nat nk /\ [alpha]_s = valpha /\
  (var_e x |--> X ** var_e y |--> Y ** var_e z |--> Z ** var_e m |--> M) s h /\ u2Z [ext]_s = Z_of_nat next /\
  (exists K, \B^next * \S_{ nk.+1 } (Z ++ [C]_s :: nil) = \S_{ next } X * \S_{ nk } Y + K * \S_{ nk } M /\
    \S_{ next } X * \S_{ nk } Y + K * \S_{ nk } M < 2 * \S_{ nk } M * \B^next) /\
  (next < nk)%nat /\ [X_]_s = X `32_ next /\
  [Y_]_s = Y `32_ 0 /\ [M_]_s = M `32_ 0 /\ [int_]_s = one32 /\
  [quot]_s = ((((X `32_ next `* Y `32_ 0) `+ zext 32 (Z `32_ 0)) `% 32) `* [alpha]_s) `% 32 /\
  \B^1 * u2Z (store.hi s `|| store.lo s) =
  u2Z [quot]_s * u2Z (M `32_ 0) + u2Z (X `32_ next) * u2Z (Y `32_ 0) + u2Z (Z `32_ 0) /\
  store.utoZ s < 2 * \B^1 - 2).

move=> s h [Z [next [HZ [r_x [r_y [r_z [r_m [r_k [r_alpha [Hmem [r_ext [Hextk' [Hone [Hinv [Hextk [r_X_ [r_Y_ [r_Z_ [r_M_ [r_int_ [Hquot [Hm2 Hm3]]]]]]]]]]]]]]]]]]]]]] .
exists Z, next; repeat Reg_upd; repeat (split; trivial).
by Assert_upd.
rewrite u2Z_concat store.hi_mflhxu_op store.lo_mflhxu_op -Hm2 store.utoZ_def.
have -> : \B^2 = \B^1 * \B^1 by rewrite -Zbeta_is_exp.
rewrite u2Z_zext Hm3 Z2uK -Zbeta1_Zpower2 //.
repeat Reg_upd; ring.

apply store.mflhxu_kbeta1_utoZ.
apply Zle_lt_trans with ((\B^1 - 1) * (\B^1 - 1) + (\B^1 - 1) * (\B^1 - 1) + (\B^1 - 1)); last by done.
Reg_upd.
rewrite Hm2; apply Zplus_le_compat.
- apply Zplus_le_compat; rewrite -u2Z_umul; by apply max_u2Z_umul.
- by apply Zle_minus_1, max_u2Z.

(** addiu t z zero16; *)

apply hoare_addiu with  (fun s h => exists Z next, size Z = nk /\
  [x]_s = vx /\ [y]_s = vy /\ [z]_s = vz /\ [m]_s = vm /\
  [one]_s = one32 /\ u2Z [k]_s = Z_of_nat nk /\ [alpha]_s = valpha /\
  (var_e x |--> X ** var_e y |--> Y ** var_e z |--> Z ** var_e m |--> M) s h /\
  u2Z [ext]_s = Z_of_nat next /\ (exists K,
    \B^next * \S_{ nk.+1 } (Z ++ [C]_s :: nil) = \S_{ next } X * \S_{ nk } Y + K * \S_{ nk } M /\
    \S_{ next } X * \S_{ nk } Y + K * \S_{ nk } M < 2 * \S_{ nk } M * \B^next) /\
  (next < nk)%nat /\ [X_]_s = X `32_ next /\
  [Y_]_s = Y `32_ 0 /\ [M_]_s = M `32_ 0 /\ [int_]_s = one32 /\
  [quot]_s = ((((X `32_ next `* Y `32_ 0) `+ zext 32 (Z `32_ 0)) `% 32) `* [alpha]_s) `% 32 /\
  \B^1 * u2Z (store.hi s `|| store.lo s) =
  u2Z [quot]_s * u2Z (M `32_ 0) + u2Z (X `32_ next) * u2Z (Y `32_ 0) + u2Z (Z `32_ 0) /\
  store.utoZ s < 2 * \B^1 - 2 /\ [t]_s = vz).

move=> s h [Z [next [HZ [r_x [r_y [r_z [r_m [Hone [r_k [r_alpha [Hmem [r_ext [Hinv [Hextk [r_X_ [r_Y_ [ [r_M_  [r_int_ [Hquot [Hm2 Hm3]]]]]]]]]]]]]]]]]]]]].

exists Z, next; repeat Reg_upd; repeat (split; trivial).
by Assert_upd.
by rewrite sext_Z2u // addi0.

(** while (bne int_ k) *)

apply (hoare_prop_m.hoare_stren (fun s h => exists Z next nint_, size Z = nk /\
  [x]_s = vx /\ [y]_s = vy /\ [z]_s = vz /\ [m]_s = vm /\
  [one]_s = one32 /\ u2Z [k]_s = Z_of_nat nk /\ [alpha]_s = valpha /\
  (var_e x |--> X ** var_e y |--> Y ** var_e z |--> Z ** var_e m |--> M) s h /\
  u2Z [ext]_s = Z_of_nat next /\ (next < nk)%nat /\ [X_]_s = X `32_ next /\
  u2Z [int_]_s = Z_of_nat nint_ /\ (1 <= nint_)%nat /\ u2Z [int_]_s <= u2Z [k]_s /\
  store.utoZ s < 2 * \B^1 - 1 /\ u2Z [t]_s = u2Z vz + 4 * (Z_of_nat nint_ - 1) /\
  (exists K, \B^next.+1 * Sum_hole nk.+1 nint_.-1 (Z ++ [C]_s :: nil) + 
    u2Z (store.hi s `|| store.lo s) * \B^(next + nint_) =
    \S_{ next } X * \S_{ nk } Y + \S_{ nint_ } Y * u2Z (X `32_ next) * \B^next +
    \S_{ nint_ } M * u2Z [quot]_s * \B^next + K * \S_{ nk } M /\
    \S_{ next } X * \S_{ nk } Y + K * \S_{ nk } M < 2 * \S_{ nk } M * \B^next))).

move=> s h [Z [next [HZ [r_x [r_y [r_z [r_m [Hone [r_k [r_alpha [Hmem [r_ext [Hinv [Hextk [r_X_ [r_Y_ [[r_M_ [r_int_ [Hquot [Hm2 [Hm3 r_t]]]]]]]]]]]]]]]]]]]]]].

exists Z, next, 1%nat; repeat (split; trivial).
by rewrite r_int_ /one32 Z2uK.
rewrite r_int_ r_k /one32 Z2uK // (_ : 1 = Z_of_nat 1) //;
  by apply/inj_le/leP; apply: leq_ltn_trans Hextk.
by apply Zlt_le_trans with (2 * \B^1 - 2).
rewrite r_t; ring.

case: Hinv => K [Hinv1 Hinv2].
exists K; split; last by done.
rewrite Sum_hole_last'; last by rewrite size_cat /= HZ addnC.
have -> : u2Z (store.hi s `|| store.lo s) * \B^(next + 1) =
  \B^next * (\B^1 * u2Z (store.hi s `|| store.lo s)) by rewrite Zbeta_is_exp; ring.
rewrite subn1.
have -> : \B^next.+1 * \S_{ nk } (List.tail (Z ++ [C]_s :: nil)) =
  \B^next * (\B^1 * \S_{ nk } (List.tail (Z ++ [C]_s :: nil))) by rewrite Zbeta_S; ring.
rewrite lSum_Zpower_Zmult Hquot.
apply trans_eq with (\B^next * 
  (\S_{ nk.+1 } (zero32 :: List.tail (Z ++ [C]_s :: nil))%list + u2Z (Z `32_ 0)) +
  \B^next * (u2Z [quot]_s * u2Z (M `32_ 0) + u2Z (X `32_ next) * u2Z (Y `32_ 0))).
- rewrite Hm2. set tmp := \S_{ _ } _. ring.
- rewrite lSum_head_swap0 tail_app; last by rewrite HZ; apply: leq_ltn_trans Hextk.
  rewrite List.app_comm_cons. 
  move: (@list_tail _ zero32 Z) => ->; last by rewrite HZ; destruct nk => //; apply lt_O_Sn.
  rewrite Hquot Hinv1 2!lSum_1 /zero32 /nth' -/(Y `32_ 0) -/(M `32_ 0); ring.

apply while.hoare_seq with (fun s h => (exists Z next nint_, size Z = nk /\
  [x]_s = vx /\ [y]_s = vy /\ [z]_s = vz /\ [m]_s = vm /\
  [one]_s = one32 /\ u2Z [k]_s = Z_of_nat nk /\ [alpha]_s = valpha /\
  (var_e x |--> X ** var_e y |--> Y ** var_e z |--> Z ** var_e m |--> M) s h /\
  u2Z [ext]_s = Z_of_nat next /\ (next < nk)%nat /\ [X_]_s = X `32_ next /\
  u2Z [int_]_s = Z_of_nat nint_ /\ (1 <= nint_)%nat /\ u2Z [int_]_s <= u2Z [k]_s /\
  store.utoZ s < 2 * \B^1 - 1 /\ u2Z [t]_s = u2Z vz + 4 * (Z_of_nat nint_ - 1) /\
  (exists K, \B^next.+1 * Sum_hole nk.+1 nint_.-1 (Z ++ [C]_s :: nil) + 
    u2Z (store.hi s `|| store.lo s) * \B^(next + nint_) =
    \S_{next} X * \S_{ nk } Y + \S_{ nint_ } Y * u2Z (X `32_ next) * \B^next +
    \S_{nint_} M * u2Z [quot]_s * \B^next + K * \S_{ nk } M /\
    \S_{next} X * \S_{ nk } Y + K * \S_{ nk } M < 2 * \S_{ nk } M * \B^next)) 
/\ ~~ (eval_b (bne int_ k) s)).

apply while.hoare_while.

(**  lwxs Y int_ y; *)

apply hoare_lwxs_back_alt'' with (fun s h => exists Z next nint_, size Z = nk /\
  [x]_s = vx /\ [y]_s = vy /\ [z]_s = vz /\ [m]_s = vm /\
  [one]_s = one32 /\ u2Z [k]_s = Z_of_nat nk /\ [alpha]_s = valpha /\
  (var_e x |--> X ** var_e y |--> Y ** var_e z |--> Z ** var_e m |--> M) s h /\
  u2Z [ext]_s = Z_of_nat next /\ (next < nk)%nat /\ (nint_ < nk)%nat /\
  [X_]_s = X `32_ next /\ u2Z [int_]_s = Z_of_nat nint_ /\ (1 <= nint_)%nat /\
  store.utoZ s < 2 * \B^1 - 1 /\ u2Z [t]_s = u2Z vz + 4 * (Z_of_nat nint_ - 1) /\
  (exists K, \B^next.+1 * Sum_hole nk.+1 nint_.-1 (Z ++ [C]_s :: nil) + 
    u2Z (store.hi s `|| store.lo s) * \B^(next + nint_) =
    \S_{ next } X * \S_{ nk } Y + \S_{ nint_ } Y * u2Z (X `32_ next) * \B^next +
    \S_{ nint_ } M * u2Z [quot]_s * \B^next + K * \S_{ nk } M /\
    \S_{ next } X * \S_{ nk } Y + K * \S_{ nk } M < 2 * \S_{ nk } M * \B^next) /\
  [Y_]_s = Y `32_ nint_).

move=> s h [ [Z [next [nint_ [HZ [r_x [r_y [r_z [r_m [Hone [r_k [r_alpha [mem [r_ext [Hextk' [r_X_ [r_int_ [Hint_ [r_int_' [Hm2 [r_t Hinv]]]]]]]]]]]]]]]]]]]] Hint_k].
rewrite /= in Hint_k; move/ZeqP in Hint_k.
have {Hint_k} Hint_k : (nint_ < nk)%nat.
  rewrite ltn_neqAle.
  apply/andP; split; [apply/eqP | ].
    rewrite r_int_ r_k in Hint_k; by apply Z_of_nat_inj_neq.
  apply/leP/Nat2Z.inj_le; by rewrite -r_int_ -r_k.

exists (Y `32_ nint_); split.
- Decompose_32 Y nint_ Y1 Y2 HlenY1 HY'; last by rewrite Hy.
  rewrite HY' (decompose_equiv _ _ _ _ _ HlenY1) in mem.
  do 2 rewrite !assert_m.conAE assert_m.conCE in mem.
  rewrite !assert_m.conAE in mem.
  move: mem; apply monotony => // h'.
  apply mapsto_ext => //.
  by rewrite /= shl_Z2u r_int_ inj_mult Zmult_comm.
- rewrite /update_store_lwxs.
  exists Z, next, nint_; repeat Reg_upd; repeat (split; trivial).
  by Assert_upd.
  
(**  lwxs Z int_ z; *)

apply hoare_lwxs_back_alt'' with (fun s h => exists Z next nint_, size Z = nk /\
  [x]_s = vx /\ [y]_s = vy /\ [z]_s = vz /\ [m]_s = vm /\
  [one]_s = one32 /\ u2Z [k]_s = Z_of_nat nk /\ [alpha]_s = valpha /\
  (var_e x |--> X ** var_e y |--> Y ** var_e z |--> Z ** var_e m |--> M) s h /\
  u2Z [ext]_s = Z_of_nat next /\ (next < nk)%nat /\ (nint_ < nk)%nat /\
  [X_]_s = X `32_ next /\ u2Z [int_]_s = Z_of_nat nint_ /\ (1 <= nint_)%nat /\
  store.utoZ s < 2 * \B^1 - 1 /\ u2Z [t]_s = u2Z [z]_s + 4 * (Z_of_nat nint_ - 1) /\
  (exists K, \B^next.+1 * Sum_hole nk.+1 nint_.-1 (Z ++ [C]_s :: nil) + 
    u2Z (store.hi s `|| store.lo s) * \B^(next + nint_) =
    \S_{next} X * \S_{ nk } Y + \S_{ nint_ } Y * u2Z (X `32_ next) * \B^next +
    \S_{nint_} M * u2Z [quot]_s * \B^next + K * \S_{ nk } M /\
    \S_{next} X * \S_{ nk } Y + K * \S_{ nk } M < 2 * \S_{ nk } M * \B^next) /\
  [Y_]_s = Y `32_ nint_ /\ [Z_]_s = Z `32_ nint_).

move=> s h [Z [next [nint_ [HZ [r_x [r_y [r_z [r_m [Hone [r_k [r_alpha [mem [r_ext [Hextk' [Hint_' [r_X_ [ r_int_' [Hint_ [Hm2 [r_t [Hinv r_Y_]]]]]]]]]]]]]]]]]]]]].

exists (Z `32_ nint_); split.
- Decompose_32 Z nint_ Z1 Z2 HlenZ1 HZ'; last by rewrite HZ.
  rewrite HZ' (decompose_equiv _ _ _ _ _ HlenZ1) in mem.
  do 3 rewrite !assert_m.conAE assert_m.conCE in mem; rewrite !assert_m.conAE in mem.
  move: mem; apply monotony => // h'.
  apply mapsto_ext => //.
  by rewrite /= shl_Z2u r_int_' inj_mult Zmult_comm.
- exists Z, next, nint_; repeat Reg_upd; repeat (split; trivial).
  by Assert_upd.
  by rewrite r_t r_z.

(** maddu X Y; *)

apply hoare_maddu with  (fun s h => exists Z next nint_, size Z = nk /\
  [x]_s = vx /\ [y]_s = vy /\ [z]_s = vz /\ [m]_s = vm /\
  [one]_s = one32 /\ u2Z [k]_s = Z_of_nat nk /\ [alpha]_s = valpha /\
  (var_e x |--> X ** var_e y |--> Y ** var_e z |--> Z ** var_e m |--> M) s h /\
  u2Z [ext]_s = Z_of_nat next /\ (next < nk)%nat /\ (nint_ < nk)%nat /\
  [X_]_s = X `32_ next /\ u2Z [int_]_s = Z_of_nat nint_ /\ (1 <= nint_)%nat /\
  store.utoZ s < \B^2 /\ u2Z [t]_s = u2Z vz + 4 * (Z_of_nat nint_ - 1) /\
  [Y_]_s = Y `32_ nint_ /\ [Z_]_s = Z `32_ nint_ /\ (exists K,
    \B^next.+1 * Sum_hole nk.+1 nint_.-1 (Z ++ [C]_s :: nil) + 
    store.utoZ s * \B^(next + nint_) = \S_{ next } X * \S_{ nk } Y +
    \S_{ nint_ } Y * u2Z (X `32_ next) * \B^next +
    \S_{ nint_ } M * u2Z [quot]_s * \B^next +
    u2Z (X `32_ next) * u2Z (Y `32_ nint_) * \B^(next + nint_) + K * \S_{ nk } M /\
    \S_{ next } X * \S_{ nk } Y + K * \S_{ nk } M < 2 * \S_{ nk } M * \B^next)).

move=> s h [Z [next [nint_ [HZ [r_x [r_y [r_z [r_m [Hone [r_k [r_alpha [Hmem [r_ext [Hextk' [Hint_' [r_X_ [r_int_' [Hint_ [Hm2 [r_t [Hinv [r_Y_ r_Z_]]]]]]]]]]]]]]]]]]]]]].
exists Z, next, nint_.
have Htmp : store.utoZ s < \B^2 * (2 ^^ store.acx_size - 1).
  apply Zlt_le_trans with (2 * \B^1 - 1); first by done.
  apply Zle_trans with (\B^2 * (2 ^^ 8 - 1)); first by done.
  apply Zmult_le_compat_l.
  apply Zminus_le_compat.
  by apply/ZleP; rewrite Zpower_2_le store.acx_size_min.
  by apply Zbeta_0'.
repeat Reg_upd; repeat (split; trivial).
by Assert_upd.
rewrite store.utoZ_maddu //.
apply Zlt_le_trans with ((\B^1 - 1) * (\B^1 - 1) + (2 * \B^1 - 1)); last by done.
apply Zplus_le_lt_compat; last by done.
by apply max_u2Z_umul.
by rewrite r_t r_z.

case: Hinv => K [Hinv1 Hinv2].
exists K; split; last by done. (* about X * Y < M *)
apply trans_eq with (\S_{next} X * \S_{nk} Y +
 \S_{nint_} Y * u2Z (X `32_ next) * \B^next + \S_{nint_} M * u2Z [quot]_s * \B^next +
 K * \S_{nk} M + u2Z (X `32_ next) * u2Z (Y `32_ nint_) * \B^(next + nint_)).
- rewrite store.utoZ_maddu // store.utoZ_def store.utoZ_acx_beta2 //.
    rewrite Z2uK // -u2Z_umul -Hinv1 u2Z_concat -Zplus_0_r_reverse r_Y_ r_X_ // -Zplus_assoc.
    f_equal.
    rewrite (Zplus_comm (u2Z (X `32_ next `* Y `32_ nint_))) Zmult_plus_distr_l Zmult_plus_distr_l.
    f_equal.
    by rewrite Zmult_plus_distr_l Zplus_comm.
  by apply Zlt_le_trans with (2 * \B^1 - 1).
- ring.

(** lwxs M int_ m; *)

apply hoare_lwxs_back_alt'' with (fun s h => exists Z next nint_, size Z = nk /\
  [x]_ s = vx /\ [y]_s = vy /\ [z]_s = vz /\ [m]_s = vm /\
  [one]_s = one32 /\ u2Z [k]_s = Z_of_nat nk /\ [alpha]_s = valpha /\
  (var_e x |--> X ** var_e y |--> Y ** var_e z |--> Z ** var_e m |--> M) s h /\
  u2Z [ext]_s =  Z_of_nat next /\ (next < nk)%nat /\ (nint_ < nk)%nat /\
  [X_]_s = X `32_ next /\ u2Z [int_]_s =  Z_of_nat nint_ /\ (1 <= nint_)%nat /\
  store.utoZ s < \B^2 /\ u2Z [t]_s = u2Z vz + 4 * (Z_of_nat nint_ - 1) /\
  [Y_]_s = Y `32_ nint_ /\ [Z_]_s = Z `32_ nint_ /\ (exists K,
    \B^next.+1 * Sum_hole nk.+1 nint_.-1 (Z ++ [C]_s :: nil) + 
    store.utoZ s * \B^(next + nint_) = \S_{ next } X * \S_{ nk } Y +
    \S_{nint_} Y * u2Z (X `32_ next) * \B^next +
    \S_{nint_} M * u2Z [quot]_s * \B^next +
    u2Z (X `32_ next) * u2Z (Y `32_ nint_) * \B^(next + nint_) + K * \S_{ nk } M /\
    \S_{next} X * \S_{ nk } Y + K * \S_{ nk } M < 2 * \S_{ nk } M * \B^next)
  /\ [M_]_s = M `32_ nint_ ).

move=> s h [Z [next [nint_ [HZ [r_x [r_y [r_z [r_m [Hone [r_k [r_alpha [mem [r_ext [Hextk' [Hint_' [r_X_ [r_int_' [Hint_ [Hm2 [r_t [r_Y_ [r_Z_ Hinv]]]]]]]]]]]]]]]]]]]]]] .

exists (M `32_ nint_); split.
- Decompose_32 M nint_ M1 M2 HlenM1 HM'; last by rewrite Hm.
  rewrite HM' (decompose_equiv _ _ _ _ _ HlenM1) in mem.
  do 4 rewrite assert_m.conCE !assert_m.conAE in mem.
  move: mem; apply monotony => // h'.
  apply mapsto_ext => //.
  by rewrite /= shl_Z2u r_int_' inj_mult Zmult_comm.
- exists Z, next, nint_; repeat Reg_upd; repeat (split; trivial).
  by Assert_upd.

(** maddu Z_ one; *)

apply hoare_maddu with (fun s h => exists Z next nint_, size Z = nk /\
  [x]_s = vx /\ [y]_s = vy /\ [z]_s = vz /\ [m]_s = vm /\
  [one]_s = one32 /\ u2Z [k]_s = Z_of_nat nk /\ [alpha]_s = valpha /\
  (var_e x |--> X ** var_e y |--> Y ** var_e z |--> Z ** var_e m |--> M) s h /\
  u2Z [ext]_s = Z_of_nat next /\ (next < nk)%nat /\ (nint_ < nk)%nat /\
  [X_]_s = X `32_ next /\ u2Z [int_]_s = Z_of_nat nint_ /\ (1 <= nint_)%nat /\
  store.utoZ s < \B^2 + \B^1 - 1 /\ u2Z [t]_s = u2Z vz + 4 * (Z_of_nat nint_ - 1) /\
  [Y_]_s = Y `32_ nint_ /\ [Z_]_s = Z `32_ nint_ /\ [M_]_s = M `32_ nint_ /\
  (exists K, \B^next.+1 * Sum_hole nk.+1 nint_.-1 (Z ++ [C]_s :: nil) + 
    store.utoZ s * \B^(next + nint_) = \S_{ next } X * \S_{ nk } Y +
    \S_{nint_} Y * u2Z (X `32_ next) * \B^next +
    \S_{nint_} M * u2Z [quot]_s * \B^next +
    u2Z (X `32_ next) * u2Z (Y `32_ nint_) * \B^(next + nint_) +
    u2Z (Z `32_ nint_) * \B^(next + nint_) + K * \S_{ nk } M /\
    \S_{next} X * \S_{nk} Y + K * \S_{nk} M < 2 * \S_{ nk } M * \B^next)).

move=> s h [Z [next [nint_ [HZ [r_x [r_y [r_z [r_m [Hone [r_k [r_alpha [Hmem [r_ext [Hextk' [Hint_' [r_X_ [r_int_' [Hint_ [Hm2 [r_t [r_Y_ [r_Z_ [Hinv r_M_]]]]]]]]]]]]]]]]]]]]]]].
exists Z, next, nint_.
have Htmp : store.utoZ s < \B^2 * (2 ^^ store.acx_size - 1).
  apply Zlt_le_trans with (\B^2 * 1); first by done.
  rewrite Zmult_1_r.
  apply Zle_trans with (\B^2 * (2 ^^ 8 - 1)); first by done.
  apply Zmult_le_compat_l.
  by apply/Zminus_le_compat/ZleP; rewrite Zpower_2_le store.acx_size_min.
  by apply Zbeta_0'.
repeat Reg_upd; repeat (split; trivial).
by Assert_upd.
rewrite store.utoZ_maddu //.
apply Zlt_le_trans with ((\B^1 - 1) + \B^2); last by done.
apply Zplus_le_lt_compat; last by done.
rewrite Hone umul_1 (u2Z_zext 32) //; by apply Zle_minus_1, max_u2Z.

case: Hinv => K [Hinv1 Hinv2].
exists K; split; last by done. (* about X * Y < M *)
rewrite store.utoZ_maddu // Hone umul_1 (u2Z_zext 32) //.
apply trans_eq with (\B^next.+1 * Sum_hole nk.+1 nint_.-1 (Z ++ [C]_s :: nil) +
  store.utoZ s * \B^(next + nint_) + u2Z (Z `32_ nint_) * \B^(next + nint_)).
by rewrite -r_Z_ {1}(Zplus_comm (u2Z [Z_]_s)) Zmult_plus_distr_l Zplus_assoc.
rewrite Hinv1; ring.

(** maddu quot M; *)

apply hoare_maddu with (fun s h => exists Z next nint_, size Z = nk /\
  [x]_s = vx /\ [y]_s = vy /\ [z]_s = vz /\ [m]_s = vm /\
  [one]_s = one32 /\ u2Z [k]_s = Z_of_nat nk /\ [alpha]_s = valpha /\
  (var_e x |--> X ** var_e y |--> Y ** var_e z |--> Z ** var_e m |--> M) s h /\
  u2Z [ext]_s = Z_of_nat next /\ (next < nk)%nat /\ (nint_ < nk)%nat /\
  [X_]_s = X `32_ next /\ u2Z [int_]_s = Z_of_nat nint_ /\ (1 <= nint_)%nat /\
  store.utoZ s < 2 * \B^2 - \B^1 /\ u2Z [t]_s = u2Z vz + 4 * (Z_of_nat nint_ - 1) /\
  [Y_]_s = Y `32_ nint_ /\ [Z_]_s = Z `32_ nint_ /\ [M_]_s = M `32_ nint_ /\
  (exists K, \B^next.+1 * Sum_hole nk.+1 nint_.-1 (Z ++ [C]_s :: nil) + 
    store.utoZ s * \B^(next + nint_) = \S_{ next } X * \S_{ nk } Y + 
    \S_{nint_} Y * u2Z (X `32_ next) * \B^next +
    \S_{nint_} M * u2Z [quot]_s * \B^next +
    u2Z (X `32_ next) * u2Z (Y `32_ nint_) * \B^(next + nint_) +
    u2Z (Z `32_ nint_) * \B^(next + nint_) +
    u2Z (M `32_ nint_) * u2Z [quot]_s * \B^(next + nint_) + K * \S_{nk} M /\
    \S_{next} X * \S_{nk} Y + K * \S_{nk} M < 2 * \S_{nk} M * \B^next)).

move=> s h [Z [next [nint_ [HZ [r_x [r_y [r_z [r_m [Hone [r_k [r_alpha [Hmem [r_ext [Hextk' [Hint_' [r_X_ [r_int_' [Hint_ [Hm2 [r_t [r_Y_ [r_Z_ [r_M_ Hinv]]]]]]]]]]]]]]]]]]]]]]].
exists Z, next, nint_.

have Htmp : store.utoZ s < \B^2 * (2 ^^ store.acx_size - 1).
  apply Zlt_le_trans with (\B^2 + \B^1 - 1); first by done.
  apply Zle_trans with (\B^2 * (2 ^^ 8 - 1)); first by done.
  by apply Zmult_le_compat_l.

repeat Reg_upd; repeat (split; trivial).
by Assert_upd.
repeat (split; trivial).

rewrite store.utoZ_maddu //.
apply Zlt_le_trans with ((\B^1 - 1) * (\B^1 - 1) + (\B^2 + \B^1 - 1)); last by done.
apply Zplus_le_lt_compat; [by apply max_u2Z_umul | done].

case: Hinv => K [Hinv1 Hinv2].
exists K; split; last by done. (* about X * Y < M *)
rewrite store.utoZ_maddu // Zmult_plus_distr_l.
apply trans_eq with (\B^next.+1 * Sum_hole nk.+1 nint_.-1 (Z ++ [C]_s :: nil) + store.utoZ s * \B^(next + nint_) + (u2Z ([quot]_s `* [M_]_s) * \B^(next + nint_))).
by rewrite (Zplus_comm (u2Z ([quot]_s `* [M_]_s) * \B^(next + nint_))) Zplus_assoc.
rewrite Hinv1 u2Z_umul r_M_; ring.

(** addiu int_ int_ one16; *)

apply hoare_addiu with (fun s h => exists Z next nint_, size Z = nk /\
  [x]_s = vx /\ [y]_s = vy /\ [z]_s = vz /\ [m]_s = vm /\
  [one]_s = one32 /\ u2Z [k]_s = Z_of_nat nk /\ [alpha]_s = valpha /\
  (var_e x |--> X ** var_e y |--> Y ** var_e z |--> Z ** var_e m |--> M) s h /\
  u2Z [ext]_s = Z_of_nat next /\ (next < nk)%nat /\ (nint_ <= nk)%nat /\
  [X_]_s = X `32_ next /\ u2Z [int_]_s = Z_of_nat nint_ /\ (2 <= nint_)%nat /\
  store.utoZ s < 2 * \B^2 - \B^1  /\ u2Z [t]_s = u2Z vz + 4 * (Z_of_nat nint_ - 2) /\
  (exists K, \B^next.+1 * Sum_hole nk.+1 nint_.-2 (Z ++ [C]_s :: nil) + 
    store.utoZ s * \B^((next + nint_).-1) = \S_{ next } X * \S_{ nk } Y +
    \S_{nint_.-1} Y * u2Z (X `32_ next) * \B^next +
    \S_{nint_.-1} M * u2Z [quot]_s * \B^next +
    u2Z (X `32_ next) * u2Z (Y `32_ nint_.-1)  * \B^((next + nint_).-1)+
    u2Z (Z `32_ nint_.-1) * \B^((next + nint_).-1) +
    u2Z (M `32_ nint_.-1) * u2Z [quot]_s * \B^((next + nint_).-1) + K * \S_{nk} M /\
    \S_{next} X * \S_{nk} Y + K * \S_{nk} M < 2 * \S_{nk} M * \B^next)).

move=> s h [Z [next [nint_ [HZ [r_x [r_y [r_z [r_m [Hone [r_k [r_alpha [Hmem [r_ext [Hextk' [Hint_' [r_X_ [r_int_' [Hint_ [Hm2 [r_t [r_Y_ [r_Z_ [r_M_ Hinv]]]]]]]]]]]]]]]]]]]]]]].

exists Z, next, nint_.+1; repeat Reg_upd; repeat (split; trivial).
by Assert_upd.
rewrite sext_Z2u // u2Z_add_Z2u //.
- by rewrite r_int_' Z_S.
- eapply Zle_lt_trans; last by apply Hnz.
  apply Zle_trans with (Z_of_nat nk).
  + rewrite r_int_' (_ : 1 = Z_of_nat 1) // -inj_plus plus_comm; apply/inj_le/leP; by rewrite plusE add1n.
  + apply Zle_plus_trans; first by apply min_u2Z.
    rewrite Zmult_comm. apply Zle_scale; by [apply Zle_0_nat | done].

rewrite Z_S [Zmult]lock /= -lock r_t; ring.

case: Hinv => K [Hinv1 Hinv2].
exists K; split; last by done. (* about X * Y < M *)
by rewrite /= {1}addnS /= Hinv1 addnS.

(** mflhxu Z_; *)

apply hoare_mflhxu with (fun s h => exists Z next nint_, size Z = nk /\
  [x]_s = vx /\ [y]_s = vy /\ [z]_s = vz /\ [m]_s = vm /\
  [one]_s = one32 /\ u2Z [k]_s = Z_of_nat nk /\ [alpha]_s = valpha /\
  (var_e x |--> X ** var_e y |--> Y ** var_e z |--> Z ** var_e m |--> M) s h /\
  u2Z [ext]_s = Z_of_nat next /\ (next < nk)%nat /\ (nint_ <= nk)%nat /\
  [X_]_s = X `32_ next /\ u2Z [int_]_s = Z_of_nat nint_ /\ (2 <= nint_)%nat /\
  store.utoZ s < 2 * \B^1 - 1 /\ u2Z [t]_s = u2Z vz + 4 * (Z_of_nat nint_-2) /\
  (exists K, \B^next.+1 * Sum_hole nk.+1 (nint_-2) (Z ++ [C]_s :: nil) + 
    u2Z (store.hi s `|| store.lo s) * \B^(next + nint_) +
    u2Z [Z_]_s * \B^((next + nint_).-1) = \S_{ next } X * \S_{ nk } Y +
    \S_{nint_.-1} Y * u2Z (X `32_ next) * \B^next +
    \S_{nint_.-1} M * u2Z [quot]_s * \B^next +
    u2Z (X `32_ next) * u2Z (Y `32_ nint_.-1)  * \B^((next + nint_).-1) +
    u2Z (Z `32_ nint_.-1) * \B^((next + nint_).-1) +
    u2Z (M `32_ nint_.-1) * u2Z [quot]_s * \B^((next + nint_).-1) + K * \S_{nk} M /\
    \S_{next} X * \S_{nk} Y + K * \S_{nk} M < 2 * \S_{nk} M * \B^next)).

move=> s h [Z [next [nint_ [HZ [r_x [r_y [r_z [r_m [Hone [r_k [r_alpha [Hmem [r_ext [Hextk' [Hint_' [r_X_ [r_int_' [Hint_ [Hm2 [r_t Hinv]]]]]]]]]]]]]]]]]]]].
exists Z, next, nint_; repeat Reg_upd; repeat (split; trivial).
by Assert_upd.

apply store.mflhxu_kbeta1_utoZ.
apply Zlt_le_trans with (2 * \B^2 - \B^1); [by Reg_upd | done].

case: Hinv => K [Hinv1 Hinv2].
exists K; split; last by done.  (* about X * Y < M *)

rewrite -Hinv1 u2Z_concat store.hi_mflhxu_op store.lo_mflhxu_op u2Z_zext store.utoZ_def -Zplus_assoc.
f_equal.
by rewrite subn2.
rewrite -Zbeta1_Zpower2 /= 3!Zmult_plus_distr_l -3!Zmult_assoc -3!Zbeta_is_exp.
rewrite 2!add1n add2n prednK; last  by rewrite addn_gt0 (ltn_trans _ Hint_) // orbT.
repeat Reg_upd; ring.

(** addiu t t four16; *)

apply hoare_addiu with (fun s h => exists Z next nint_, size Z = nk /\
  [x]_s = vx /\ [y]_s = vy /\ [z]_s = vz /\ [m]_s = vm /\
  [one]_s = one32 /\ u2Z [k]_s = Z_of_nat nk /\ [alpha]_s = valpha /\
  (var_e x |--> X ** var_e y |--> Y ** var_e z |--> Z ** (var_e m |--> M)) s h /\
  u2Z [ext]_s = Z_of_nat next /\ (next < nk)%nat /\ (nint_ <= nk)%nat /\
  [X_]_s = X `32_ next /\ u2Z [int_]_s = Z_of_nat nint_ /\ (2 <= nint_)%nat /\
  store.utoZ s < 2 * \B^1 - 1 /\ u2Z [t]_s = u2Z vz + 4 * (Z_of_nat nint_ - 1) /\
  (exists K, \B^next.+1 * Sum_hole nk.+1 nint_.-2 (Z ++ [C]_s :: nil) + 
    u2Z (store.hi s `|| store.lo s) * \B^(next + nint_) +
    u2Z [Z_]_s * \B^((next + nint_).-1) = \S_{ next } X * \S_{ nk } Y +
    \S_{nint_.-1} Y * u2Z (X `32_ next) * \B^next +
    \S_{nint_.-1} M * u2Z [quot]_s * \B^next +
    u2Z (X `32_ next) * u2Z (Y `32_ nint_.-1)  * \B^((next + nint_).-1) +
    u2Z (Z `32_ nint_.-1) * \B^((next + nint_).-1) +
    u2Z (M `32_ nint_.-1) * u2Z [quot]_s * \B^((next + nint_).-1) + K * \S_{nk} M /\
    \S_{next} X * \S_{nk} Y + K * \S_{nk} M < 2 * \S_{nk} M * \B^next)).

move=> s h [Z [next [nint_ [HZ [r_x [r_y [r_z [r_m [Hone [r_k [r_alpha [Hmem [r_ext [Hextk' [Hint_' [r_X_ [r_int_' [Hint_ [Hm2 [r_t Hinv]]]]]]]]]]]]]]]]]]]].

exists Z, next, nint_; repeat Reg_upd; repeat (split; trivial).
by Assert_upd.
rewrite sext_Z2u // u2Z_add_Z2u //.
- rewrite r_t; ring.
- rewrite r_t -Zbeta1_Zpower2.
  eapply Zle_lt_trans; last by apply Hnz.
  rewrite -Zplus_assoc; apply Zplus_le_compat_l.
  rewrite -{2}(Zmult_1_r 4) -Zmult_plus_distr_r; apply Zmult_le_compat_l; last by done.
  rewrite -Zplus_assoc; apply Zle_plus_trans_L => //; by apply/inj_le/leP.
  by rewrite subn2 in Hinv.

(** sw Z_ mfour16 t *)

apply hoare_sw_back'.

move=> s h [Z [next [nint_ [HZ [r_x [r_y [r_z [r_m [Hone [r_k [r_alpha [Hmem [r_ext [Hextk' [Hint_' [r_X_ [r_int_' [ Hint_ [Hm2 [r_t Hinv]]]]]]]]]]]]]]]]]]]].

exists (int_e (Z `32_ nint_.-2)).

have Htmp: [z]_s `+ Z2u 32 (Z_of_nat (4 * nint_.-2)) = [t]_s `+ sext 16 mfour16.
  rewrite /mfour16 sext_Z2s // r_z; apply u2Z_inj.
  rewrite u2Z_add_Z2s //.
  rewrite u2Z_add_Z_of_nat.
  rewrite r_t inj_mult -subn2 inj_minus1 //; last by apply/leP.
  ring.
  rewrite -subn2 inj_mult inj_minus1 //; last by apply/leP.
  rewrite -Zbeta1_Zpower2 [Z_of_nat _]/=.
  eapply Zle_lt_trans; last by apply Hnz.
  apply Zplus_le_compat_l, Zmult_le_compat_l; last by done.
  rewrite -inj_minus1; last by apply/leP.
  by apply/inj_le/leP; rewrite minusE leq_subLR (leq_trans Hint_') // leq_addl.
  rewrite r_t -Zplus_assoc (_ : -4 = 4 * ( -1 )) // -Zmult_plus_distr_r.
  apply Zle_plus_trans; first by apply min_u2Z.
  rewrite Zmult_comm; apply Zmult_gt_0_le_0_compat; first by done.
  rewrite -Zplus_assoc /= (_ : -2 = - Z_of_nat 2) //.
  by apply/Zle_left/inj_le/leP.

Decompose_32 Z (nint_.-2) Z1 Z2 HlenZ1 HZ'; last first.
  by rewrite HZ (leq_trans _ Hint_') // -subn2 -subSn // subn2 leq_pred.
rewrite HZ' (decompose_equiv _ _ _ _ _ HlenZ1) in Hmem.
do 3 rewrite !assert_m.conAE assert_m.conCE in Hmem; rewrite !assert_m.conAE in Hmem.
move: Hmem; apply monotony => // ht.
by apply mapsto_ext.

apply currying => h' H'; simpl app in H'.

exists (upd_nth Z nint_.-2 [Z_]_s), next, nint_; repeat (split; trivial).
by apply size_upd_nth.
rewrite HZ' upd_nth_cat HlenZ1 // subnn /= (decompose_equiv _ _ _ _ _ HlenZ1).
rewrite cat0s in H'.
assoc_comm H'.
move: H'.
by apply mapsto_ext.
by rewrite (ltn_trans _ Hint_).
rewrite r_int_' r_k; by apply/inj_le/leP.
case: Hinv => K [Hinv1 Hinv2].
exists K; split; last by done. (* about X * Y < M *)

(* technicalities:

   at this point, we have an hypothesis about K with
   ( z_0 ... z_{nint-2} ... z_{k-1} ++ C::nil )\{nint-2}

   whereas in the goal, we have
   ( z_0 ...     Z      ... z_{k-1} ++ C::nil )\{nint-1}

   we will:
   1. coax the hypothesis and the goal to have the same Sum_hole predicate
      1.a. add z_{nint-1} to the goal
      1.b. use the Sum_hole_shift property to change the Sum_hole predicate in the goal
   2. check that the goal can be obtained from the hypo by rewriting
 *)

(* 1. *)

(* 1.a. *)

apply Zplus_reg_l with (\B^next.+1 * u2Z (Z `32_ nint_.-1) * \B^nint_.-2).
rewrite Zplus_assoc -Zmult_assoc -(Zmult_plus_distr_r (\B^next.+1)).

(* 1.b. *)

rewrite (Zplus_comm (u2Z (Z `32_ nint_.-1) * \B^nint_.-2)).
have -> : Sum_hole nk.+1 nint_.-1 (upd_nth Z nint_.-2 [Z_]_s ++ [C]_s :: nil) +
    u2Z (Z `32_ nint_.-1) * \B^nint_.-2 =
    Sum_hole nk.+1 nint_.-2 (Z ++ [C]_s :: nil) + u2Z [Z_]_s * \B^nint_.-2.
  have H_ : size (upd_nth Z nint_.-2 [Z_]_s ++ [C]_s :: nil) = nk.+1.
    by rewrite size_cat (@size_upd_nth _ nk) // addn1.
  have H1__ : (nint_.-1 < nk)%nat by rewrite prednK // (ltn_trans _ Hint_).
  have H1_ : (nint_.-1 < nk.+1)%nat by apply ltnW.
  have H1___ : (nint_.-2 < nk)%nat by rewrite prednK // -ltnS // prednK // (ltn_trans _ Hint_).
  move/(Sum_hole_shift _ _ H_ _) : H1_ => {H_}H_.
  do 2 rewrite nth_cat (@size_upd_nth _ nk) // in H_.
  rewrite H1__ subn1 H1___ nth_upd_nth' in H_; last first.
    destruct nint_; first by done.
    destruct nint_; first by done.
    destruct nint_ as [|nint_]; first by done.
    apply/eqP; by rewrite neq_ltn /= ltnSn orbT.
  rewrite nth_upd_nth in H_; last by rewrite HZ.
  have [->//] : Sum_hole nk.+1 nint_.-2 (Z ++ [C]_s :: nil) =
                Sum_hole nk.+1 nint_.-2 (upd_nth Z nint_.-2 [Z_]_s ++ [C]_s :: nil).
  rewrite -upd_nth_cat'; last by rewrite HZ.
  rewrite -Sum_hole_upd_nth //.
  by rewrite size_cat addn1 /= HZ.
  by rewrite mulnC in H_.
(* 2. *)

have Htmp2 : (next.+1 + nint_.-2 = (next + nint_).-1)%nat.
  rewrite addSnnS prednK //; last by rewrite -ltnS // prednK // (ltn_trans _ Hint_).
  rewrite -[RHS]subn1 -addnBA //; last by rewrite (ltn_trans _ Hint_).
  by rewrite subn1.

apply trans_eq with (\B^next.+1 * Sum_hole nk.+1 nint_.-2 (Z ++ [C]_s :: nil) +
 u2Z (store.hi s `|| store.lo s) * \B^(next + nint_) +
  u2Z [Z_]_s * (\B^next.+1 * \B^nint_.-2)).
ring.

rewrite -Zbeta_is_exp Htmp2 {}Hinv1.

apply trans_eq with ((\S_{nint_.-1} Y * u2Z (X `32_ next) * \B^next +
 u2Z (X `32_ next) * u2Z (Y `32_ nint_.-1) * \B^((next + nint_).-1)) +
 (\S_{nint_.-1} M * u2Z [quot]_s * \B^next +
 u2Z (M `32_ nint_.-1) * u2Z [quot]_s * \B^((next + nint_).-1)) 
 + K * \S_{nk} M
 + \S_{next} X * \S_{nk} Y
 + u2Z (Z `32_ nint_.-1) * \B^((next + nint_).-1)).
ring.

apply trans_eq with (\S_{ nint_ } Y * u2Z (X `32_ next) * \B^next +
 \S_{ nint_ } M * u2Z [quot]_s * \B^next
 + K * \S_{ nk } M + \S_{ next } X * \S_{ nk } Y
 + (u2Z (Z `32_ nint_.-1) * (\B^next.+1 * \B^nint_.-2))); last by ring.

f_equal; last by rewrite -Zbeta_is_exp Htmp2.

do 3 f_equal.
apply trans_eq with ((\S_{nint_.-1} Y + u2Z (Y `32_ nint_.-1) *
  \B^(nint_.-1)) * (u2Z (X `32_ next) * \B^next)).
- rewrite -(subn1 (next + nint_)) -addnBA; last by apply: leq_trans Hint_.
  rewrite subn1 Zbeta_is_exp; ring.
- rewrite (Zmult_comm _ (\B^nint_.-1)) {1}/Zbeta -lSum_remove_last.
- rewrite prednK //.
  by rewrite Zmult_assoc.
  by apply: leq_trans Hint_.

apply trans_eq with ((\S_{nint_.-1} M + u2Z (M `32_ nint_.-1) *
  \B^nint_.-1) * (u2Z [quot]_s * \B^next)).
- rewrite -(subn1 (next + nint_)) -addnBA; last by apply: leq_trans Hint_.
  rewrite subn1 Zbeta_is_exp; ring.
- rewrite (Zmult_comm _ (\B^nint_.-1)) -lSum_remove_last prednK //.
  by rewrite Zmult_assoc.
  by apply: leq_trans Hint_.

(** maddu C one; *)

apply hoare_maddu with (fun s h => exists Z next, size Z = nk /\
  [x]_s = vx /\ [y]_s = vy /\ [z]_s = vz /\ [m]_s = vm /\
  [one]_s = one32 /\ u2Z [k]_s = Z_of_nat nk /\ [alpha]_s = valpha /\
  (var_e x |--> X ** var_e y |--> Y ** var_e z |--> Z ** var_e m |--> M) s h /\
  u2Z [ext]_s = Z_of_nat next /\ (next < nk)%nat /\
  [X_]_s = X `32_ next /\ u2Z [int_]_s = Z_of_nat nk /\
  store.utoZ s < 3 * \B^1 - 2 /\ u2Z [t]_s = u2Z vz + 4 * (Z_of_nat nk - 1) /\
  (exists K, \B^next.+1 * \S_{nk.-1} Z + store.utoZ s * \B^(next + nk) =
    \S_{next} X * \S_{ nk } Y + \S_{ nk } Y * u2Z (X `32_ next) * \B^next +
    \S_{nk} M * u2Z [quot]_s * \B^next + K * \S_{ nk } M /\
    \S_{next} X * \S_{ nk } Y + K * \S_{ nk } M < 2 * \S_{ nk } M * \B^next)).

move=> s h [[Z [next [nint_ [HZ [r_x [r_y [r_z [r_m [Hone [r_k [r_alpha [Hmem [r_ext [Hextk' [ Hint_' [r_int_' [r_X_ [Hint_ [Hm2 [r_t Hinv]]]]]]]]]]]]]]]]]]]] Hint_k].

rewrite /= r_int_' r_k in Hint_k. move/negPn/ZeqP/Z_of_nat_inj in Hint_k; subst nint_.

exists Z, next.

have Htmp : store.utoZ s < \B^2 * (2 ^^ store.acx_size - 1).
  apply Zlt_le_trans with (2 * \B^1 - 1); first by done.
  by apply Zle_trans with (\B^2 * (2 ^^ 8 - 1)).

repeat Reg_upd; repeat (split; trivial).
by Assert_upd.

rewrite store.utoZ_maddu // Hone umul_1 (@u2Z_zext 32) //.
apply Zlt_le_trans with ((\B^1 - 1) + (2 * \B^1 - 1)); last by done.
apply Zplus_le_lt_compat; last by done.
by apply Zle_minus_1, max_u2Z.

case: Hinv => K [Hinv1 Hinv2].
exists K; split; last by done. (* about X * Y < M *)
rewrite -Hinv1 store.utoZ_maddu // Hone umul_1 (@u2Z_zext 32) Zmult_plus_distr_l Zplus_assoc /Sum_hole //.
rewrite subn1 store.utoZ_def store.utoZ_acx_beta2; last by apply Zlt_le_trans with (2 * \B^1 - 1).
rewrite Z2uK; last by split => //; apply Zpower_0.
rewrite Zmult_0_l -Zplus_0_r_reverse u2Z_concat.
have [Z' [tl H ] ] : exists Z' tl, Z = Z' ++ tl :: nil.
  clear -HZ r_X_.
  case/lastP : Z HZ r_X_ => /=; first by move=> <-.  
  move=> h t _ _; by exists h, t; rewrite -cats1.
rewrite {2}H -catA.
have H0 : size Z' = nk.-1.
  rewrite H size_cat /= addn1 in HZ; by rewrite -HZ.
rewrite (idel_app _ Z' H0 _ tl ([C]_s :: nil)) //.
move: (lSum_remove_last _ (Z' ++ [C]_s :: nil) nk.-1) => H1.
rewrite -(@lSum_beyond _ nk.-1) in H1; last by done.
rewrite nth_cat H0 ltnn subnn [nth _ _ _]/= in H1.
rewrite prednK // in H1.
rewrite H1 (Zmult_plus_distr_r (\B^next.+1)).
have <- : \S_{nk.-1} Z = \S_{nk.-1} Z' by rewrite H -lSum_beyond.
rewrite -ZbetaE Zmult_assoc -Zbeta_is_exp.
rewrite addSnnS prednK //.
by rewrite (Zmult_comm (\B^(next + nk))) (Zplus_comm (u2Z (store.lo s))).

(** mflhxu Z_; *)

apply hoare_mflhxu with (fun s h => exists Z next, size Z = nk /\
  [x]_s = vx /\ [y]_s = vy /\ [z]_s = vz /\ [m]_s = vm /\
  [one]_s = one32 /\ u2Z [k]_s = Z_of_nat nk /\ [alpha]_s = valpha /\
  (var_e x |--> X ** var_e y |--> Y ** var_e z |--> Z ** var_e m |--> M) s h /\
  u2Z [ext]_s = Z_of_nat next /\ (next < nk)%nat /\ [X_]_s = X `32_ next /\
  store.utoZ s < 3  /\ u2Z [t]_s = u2Z vz + 4 * (Z_of_nat nk - 1) /\ 
  (exists K, \B^next.+1 * \S_{nk.-1} Z + u2Z (store.lo s) * \B^(next + nk + 1) +
    u2Z [Z_]_s * \B^(next + nk) = \S_{ next } X * \S_{ nk } Y +
    \S_{ nk } Y * u2Z (X `32_ next) * \B^next +
    \S_{ nk } M * u2Z [quot]_s * \B^next + K * \S_{ nk } M /\
    \S_{ next } X * \S_{ nk } Y + K * \S_{ nk } M < 2 * \S_{ nk } M * \B^next)).

move=> s h [Z [next [HZ [r_x [r_y [r_z [r_m [Hone [r_k [r_alpha [Hmem [r_ext [Hextk' [r_X_ [r_int_ [Hm2 [r_t Hinv]]]]]]]]]]]]]]]]].

exists Z, next; repeat Reg_upd; repeat (split; trivial).
by Assert_upd.
apply store.mflhxu_kbeta1_utoZ.
apply Zlt_trans with (3 * \B^1 - 2); [by Reg_upd | done].
case: Hinv => K [Hinv1 Hinv2].
exists K; split; last by done. (* about X * Y < M *)
rewrite -Hinv1 store.lo_mflhxu_op store.utoZ_def store.utoZ_acx_beta2; last by apply Zlt_le_trans with (3 * \B^1 -2).
rewrite Z2uK; last by split; [apply Zle_refl | apply Zpower_0].
rewrite Zmult_0_l -Zplus_0_r_reverse Zmult_plus_distr_l -Zmult_assoc -Zbeta_is_exp (addnC 1%nat).
Reg_upd; ring.

(** addiu ext ext one16; *)

apply hoare_addiu with (fun s h => exists Z, exists next, size Z = nk /\
  [x]_s = vx /\ [y]_s = vy /\ [z]_s = vz /\ [m]_s = vm /\
  [one]_s = one32 /\ u2Z [k]_s = Z_of_nat nk /\ [alpha]_s = valpha /\
  (var_e x |--> X ** var_e y |--> Y ** var_e z |--> Z ** var_e m |--> M) s h /\
  u2Z [ext]_s = Z_of_nat next /\ (1 <= next <= nk)%nat /\
  store.utoZ s < 3 /\ u2Z [t]_s = u2Z vz + 4 * (Z_of_nat nk - 1) /\ 
  (exists K, \B^next * \S_{ nk.-1 } Z + u2Z (store.lo s) * \B^(next + nk) +
    u2Z [Z_]_s * \B^((next + nk).-1) = \S_{ next.-1 } X * \S_{ nk } Y +
    \S_{ nk } Y * u2Z (X `32_ next.-1) * \B^next.-1 +
    \S_{ nk } M * u2Z [quot]_s * \B^next.-1 + K * \S_{ nk } M /\
    \S_{ next } X * \S_{ nk } Y + K * \S_{ nk } M + (\B^1 - 1) * (\B^next.-1 * \S_{ nk } M) < 2 * \S_{ nk } M * \B^next)).

move=> s h [Z [next [HZ [r_x [r_y [r_z [r_m [Hone [r_k [r_alpha [Hmem [r_ext [Hextk' [r_X_ [Hm2 [r_t Hinv]]]]]]]]]]]]]]]].

exists Z, next.+1; repeat Reg_upd; repeat (split; trivial).
by Assert_upd.

rewrite sext_Z2u // u2Z_add_Z2u // r_ext.
by rewrite Z_S.
apply Zle_lt_trans with (Z_of_nat nk).
rewrite (_ : 1 = Z_of_nat 1) // -inj_plus plusE addn1; by apply/inj_le/leP.
rewrite -r_k; by apply max_u2Z.

case: Hinv => K [Hinv1 Hinv2].
exists K; split.
- by rewrite -Hinv1 addSn addnS addn0.
- (* about X * Y < M *)
  rewrite lSum_remove_last Zmult_plus_distr_l.
  apply Zlt_le_trans with (2 * \S_{ nk } M * \B^next + u2Z (X `32_ next) * \B^next * \S_{ nk } Y +
    (\B^1 - 1) * \B^(next.+1.-1) * \S_{ nk } M).
  apply Zplus_lt_le_compat; last by rewrite -Zmult_assoc; apply Zle_refl.
  rewrite -ZbetaE (Zmult_comm (\B^next)) /zero32 Zplus_comm Zplus_assoc; 
    by apply Zplus_lt_le_compat; [rewrite Zplus_comm | apply Zle_refl].
  apply Zle_trans with (2 * \S_{ nk } M * \B^next + (\B^1 - 1) * \B^next * \S_{ nk } Y + 
    (\B^1 - 1) * \B^(next.+1.-1) * \S_{ nk } M).
  apply Zplus_le_compat_r, Zplus_le_compat_l.
  rewrite -2!Zmult_assoc. 
  apply Zmult_le_compat_r.
  apply Zle_minus_1; by apply max_u2Z.
  apply Zmult_le_0_compat; by [apply Zbeta_0' | apply min_lSum].
  rewrite -Zplus_assoc.
  apply Zle_trans with (2 * \S_{ nk } M * \B^next +
    ((\B^1 - 1) * \B^next * \S_{ nk } M + (\B^1 - 1) * \B^next * \S_{ nk } M)).
  apply Zplus_le_compat_l, Zplus_le_compat_r, Zmult_le_compat_l.
  by apply Zlt_le_weak.
  apply Zmult_le_0_compat; [done | by apply Zbeta_0'].
  suff: 2 * \S_{ nk } M * \B^next + ((\B^1 - 1) * \B^next * \S_{ nk } M + (\B^1 - 1) * \B^next * \S_{ nk } M) 
    = 2 * \S_{ nk } M * \B^next.+1 by move=> ->; apply Zle_refl.
  rewrite (Zbeta_S next); ring.

(** sw Z_ zero16 t; *)

apply hoare_sw_back'' with (fun s h => exists Z next, size Z = nk /\
  [x]_s = vx /\ [y]_s = vy /\ [z]_s = vz /\ [m]_s = vm /\
  [one]_s = one32 /\ u2Z [k]_s = Z_of_nat nk /\ [alpha]_s = valpha /\
  (var_e x |--> X ** var_e y |--> Y ** var_e z |--> Z ** var_e m |--> M) s h /\
  u2Z [ext]_s = Z_of_nat next /\ (1 <= next <= nk)%nat /\
  store.utoZ s < 3 /\ u2Z [t]_s = u2Z vz + 4 * (Z_of_nat nk - 1) /\ 
  (exists K, \B^next * \S_{ nk } Z + u2Z (store.lo s) * \B^(next + nk) =
    \S_{next.-1} X * \S_{ nk } Y + \S_{ nk } Y * u2Z (X `32_ next.-1) * \B^next.-1 +
    \S_{nk} M * u2Z [quot]_s * \B^(next - 1) + K * \S_{ nk } M /\
    \S_{next} X * \S_{ nk } Y + K * \S_{ nk } M + (\B^1 - 1) * (\B^next.-1 * \S_{ nk } M) < 2 * \S_{ nk } M * \B^next)).

move=> s h [Z [next [HZ [r_x [r_y [r_z [r_m [Hone [rk [r_alpha [Hmem [r_ext [Hextk' [Hm2 [r_t Hinv]]]]]]]]]]]]]]].

have [lst1 [last H0]] : exists Z' tl, Z = Z' ++ tl :: nil.
  clear -HZ Hextk'.  
  case/lastP : Z HZ Hextk' => [<-|].
    rewrite leqn0; case/andP => H1 /eqP H2; by rewrite H2 in H1.
  move=> h t _ _; by exists h, t; rewrite cats1.
have Hlenlst1 : size lst1 = nk.-1 by rewrite -HZ H0 size_cat /= addn1.
have Htmp : [ var_e t \+ int_e (sext 16 zero16) ]e_ s = [ var_e z \+ int_e (Z2u 32 (Z_of_nat (4 * (nk - 1)))) ]e_ s.
  rewrite /= sext_0 addi0; apply u2Z_inj.
  rewrite u2Z_add_Z_of_nat.
  rewrite inj_mult inj_minus1; last by destruct nk => //; by apply le_n_S, le_O_n.
  rewrite r_t r_z; simpl Z_of_nat; ring.
  rewrite inj_mult inj_minus1; last by destruct nk => //; by apply le_n_S, le_O_n.
  rewrite r_z; simpl Z_of_nat.
  eapply Zle_lt_trans; last by apply Hnz.
  apply Zplus_le_compat_l.
  apply Zmult_le_compat_l; last by done.
  rewrite (_ : 1 = Z_of_nat 1) // -inj_minus1 //; last first.
    by apply/leP; apply leq_trans with next; case/andP: Hextk'.
  apply/inj_le/leP; by rewrite minusE subn1 leq_pred.

exists (int_e last).

rewrite H0 (decompose_equiv _ _ _ _ _ Hlenlst1) in Hmem.
do 3 rewrite !assert_m.conAE assert_m.conCE in Hmem.
rewrite !assert_m.conAE in Hmem.
move: Hmem; apply monotony => // ht.
rewrite -subn1; by apply mapsto_ext.
apply currying => h' H'; simpl assert_m.mapstos in H'.

exists (upd_nth Z nk.-1 [Z_]_s), next.

repeat (split; trivial).
by apply size_upd_nth.
rewrite H0 upd_nth_cat Hlenlst1 // subnn /= (_ : lst1 ++ [Z_]_s :: nil = lst1 ++ ([Z_]_s :: nil) ++ nil) // (decompose_equiv _ _ _ _ _ Hlenlst1).
simpl assert_m.mapstos.
assoc_comm H'.
by move: H'; rewrite -subn1; apply mapsto_ext.

case: Hinv => K [Hinv1 Hinv2].
exists K; split; last by done. (* about X * Y < M *)
rewrite subn1 -Hinv1.
have -> : upd_nth Z nk.-1 [Z_]_s = lst1 ++ [Z_]_s :: nil.
  rewrite H0 upd_nth_cat /=.
  by rewrite Hlenlst1 subnn.
  by rewrite Hlenlst1.
rewrite (lSum_cut_last nk) //; last by rewrite size_cat /= -HZ H0 size_cat.
rewrite Zmult_plus_distr_r Zmult_assoc -Zbeta_is_exp !subn1.
rewrite H0 -(lSum_beyond 32 nk.-1 lst1) //.
rewrite -(subn1 (next + nk)) -addnBA //.
rewrite subn1; ring.
rewrite (@leq_trans next) //.
by case/andP : Hextk'.
by case/andP : Hextk'.

(** mflhxu C *)

apply hoare_mflhxu'.

move=> s h [Z [next [HZ [r_x [r_y [r_z [r_m [Hone [r_k [r_alpha [Hmem [r_ext [Hextk' [Hm2 [r_t Hinv]]]]]]]]]]]]]]].

exists Z, next; repeat Reg_upd; repeat (split; trivial).
by Assert_upd.

rewrite r_ext r_k.

apply/inj_le/leP; by case/andP: Hextk'.

rewrite -subn1 inj_minus1 //; destruct nk => //; by apply le_n_S, le_O_n.

case: Hinv => K [Hinv1 Hinv2].
exists (K + u2Z [quot]_s * \B^next.-1); split.
- rewrite lSum_cut_last; last by rewrite size_cat /= HZ addnC.
  rewrite subn1 Zmult_plus_distr_r Zmult_assoc -Zbeta_is_exp (Zmult_comm (\B^(next + nk))) Hinv1.
  have <- : \S_{next.-1} X + u2Z (X `32_ next.-1) * \B^next.-1 = \S_{ next } X.
    rewrite -(Zmult_comm (\B^next.-1)) /Zbeta -lSum_remove_last //.
    rewrite prednK //.
    by case/andP: Hextk'.
  rewrite subn1; ring.
(* about X * Y < M *)
- apply Zle_lt_trans with (\S_{ next } X * \S_{ nk } Y + K * \S_{ nk } M + ((\B^1 - 1) * \B^next.-1 * \S_{ nk } M)); last by rewrite -Zmult_assoc.
  rewrite Zmult_plus_distr_l Zplus_assoc.
  apply Zplus_le_compat_l, Zmult_le_compat_r; last by apply min_lSum.
  apply Zmult_le_compat_r; last by apply Zbeta_0'.
  apply Zle_minus_1; by apply max_u2Z.
Qed.
