(* seplog (c) AIST 2005-2015. R. Affeldt, N. Marti, et al. GNU GPLv3. *)
From mathcomp Require Import ssreflect ssrbool eqtype.
Require Import ssrZ ZArith_ext seq_ext.

Module Type IntegralType.

Parameter t : eqType.
Axiom t_dec : forall (x y : t), {x = y} + {x <> y}.
Parameter nat2t : nat -> t.

Parameter t_plus : t -> t -> t.
Axiom t_plus_assoc : forall a b c, t_plus a (t_plus b c) = t_plus (t_plus a b) c.
Axiom t_plus_comm : forall n m, t_plus n m = t_plus m n.
Axiom t_plus_0 : forall a, t_plus a (nat2t 0) = a.
Axiom t_plus_1 : forall a, t_plus (nat2t 1) (nat2t a) = nat2t (S a).
Parameter t_minus : t -> t -> t.
Parameter t_mult : t -> t -> t.
Parameter t_div : t -> t -> t.
Parameter t_mod : t -> t -> t.
Parameter t_valabs : t -> t.

Parameter t_ge : t -> t -> Prop.
Parameter t_gt : t -> t -> Prop.
Parameter t_geb : t -> t -> bool.
Parameter t_gtb : t -> t -> bool.

Axiom t_geb_true : forall a b : t, t_geb a b = true -> t_ge a b.
Axiom t_geb_true' : forall a b : t, t_ge a b -> t_geb a b = true.
Axiom t_gtb_true : forall a b : t, t_gtb a b = true -> t_gt a b.
Axiom t_gtb_true' : forall a b : t, t_gt a b -> t_gtb a b = true.

Parameter t_eqb : t -> t -> bool.
Axiom t_eqb_eq : forall x y : t, t_eqb x y = true -> x = y.
Axiom t_eqb_refl : forall x : t, t_eqb x x = true.
Axiom t_eqb_false : forall x y : t, t_eqb x y = false <-> x <> y.
Axiom t_eqb_false' : forall x y : t, x <> y -> t_eqb x y = false.

Axiom nat2t_inj : forall a b, nat2t a = nat2t b -> a = b.

Axiom tP : Equality.axiom t_eqb.

Parameter t_non_zero : Type.
Parameter t_non_zero2t : t_non_zero -> t.
Axiom t_non_zero2t_0 : forall x, t_non_zero2t x <> nat2t 0.
Axiom t2t_non_zero : forall (i : t), i <> nat2t O -> t_non_zero.
Axiom t_non_zero2t_t2t_non_zero : forall z (H : z <> nat2t O), t_non_zero2t (t2t_non_zero z H) = z.
Axiom t_non_zero2t_inj : forall x y, t_non_zero2t x = t_non_zero2t y -> x = y.
Parameter t_non_zero_eqb : t_non_zero -> t_non_zero -> bool.
Axiom t_non_zeroP : Equality.axiom t_non_zero_eqb.

Axiom t_ge_refl : forall a, t_ge a a.

Axiom t_gt_t_ge : forall a b, (~~ t_gtb a b) = t_geb b a.

Axiom t_gtb_def : forall a b, t_gtb a b = t_geb a b && (a != b).

Axiom t_gt_weak : forall a b, t_gt a b -> t_ge a b.

Axiom t_ge_trans : forall a b c, t_ge a b -> t_ge b c -> t_ge a c.

Axiom t_gt_1 : forall n, t_gt (t_plus (nat2t 1) n) n.

End IntegralType.

Module IntegralType_Prop (X : IntegralType).

Import X.

Lemma t_inj_plus : forall n m : nat, nat2t (n + m) = t_plus (nat2t n) (nat2t m).
Proof.
elim=> [m|n IH m].
by rewrite t_plus_comm t_plus_0.
rewrite addSn -(t_plus_1 n) -(t_plus_1 (_ + _)) -t_plus_assoc; f_equal.
by apply IH.
Qed.

(* TODO: move to Max_ext.v ? *)
Definition my_max (a b : t) := if t_geb a b then a else b.

Lemma le_my_max_r : forall a b, t_ge (my_max a b) b.
Proof.
move=> a b.
rewrite /my_max.
move H : (t_geb _ _) => [].
by apply t_geb_true.
by apply t_ge_refl.
Qed.

Lemma le_my_max_l : forall a b, t_ge (my_max b a) b.
Proof.
move=> a b.
rewrite /my_max.
move H : (t_geb _ _) => [].
by apply t_ge_refl.
move/negbT : H => H.
rewrite -t_gt_t_ge in H.
move/negPn : H => H.
apply t_gt_weak.
by apply t_gtb_true.
Qed.

Lemma my_ge_gt_dec : forall a b, t_ge a b \/ t_gt b a.
Proof.
move=> a b.
move H : (t_geb a b) => [|].
left.
by apply t_geb_true.
right.
apply t_gtb_true.
apply/negPn.
move/negbT : H.
apply contra.
move=> H.
apply t_geb_true'.
rewrite t_gt_t_ge in H.
by apply t_geb_true.
Qed.

Lemma t_gt_trans : forall a b c, t_gt a b -> t_gt b c -> t_gt a c.
Proof.
move=> a b c b_a.
move/t_gtb_true'.
move/negPn.
rewrite t_gt_t_ge => c_b.
apply t_gtb_true.
apply/negPn.
rewrite t_gt_t_ge.
move: c_b.
apply contra => c_a.
apply t_gt_weak in b_a.
apply t_geb_true'.
apply t_ge_trans with a => //.
by apply t_geb_true.
Qed.

Lemma my_max_assoc : forall a b c, my_max a (my_max b c) = my_max (my_max a b) c.
Proof.
move=> a b c.
rewrite /my_max.
move b_c : (t_geb b c) => [|].
move a_b : (t_geb a b) => [|].
move a_c : (t_geb a c) => [|] //.
apply t_geb_true in b_c.
apply t_geb_true in a_b.
move: (t_ge_trans _ _ _ a_b b_c).
move/t_geb_true'.
by rewrite a_c.
by rewrite b_c.
move a_c : (t_geb a c) => [|].
move a_b : (t_geb a b) => [|].
by rewrite a_c.
move/negbT : b_c.
rewrite -t_gt_t_ge.
move/negPn => c_b.
move/negbT : a_b.
rewrite -t_gt_t_ge.
move/negPn => b_a.
have a_c2 : t_gtb c a = true.
  apply t_gtb_true'.
  apply t_gt_trans with b => //.
  by apply t_gtb_true.
  by apply t_gtb_true.
rewrite -t_gt_t_ge in a_c.
by rewrite a_c2 in a_c.
move a_b : (t_geb a b) => [|].
by rewrite a_c.
by rewrite b_c.
Qed.

Definition my_max_list l def := List.fold_left my_max l def.

Lemma my_max_list_prop1 : forall tl def x, t_ge (my_max_list tl (my_max def x)) x.
Proof.
elim => [def x /= | hd tl IH def x /=].
by apply le_my_max_r.
case: (my_ge_gt_dec hd x) => X.
by apply t_ge_trans with hd.
rewrite -my_max_assoc.
apply t_ge_trans with (my_max x hd) => //.
by apply le_my_max_l.
Qed.

Lemma my_max_list_prop : forall l def x, List.In x l -> t_ge (my_max_list l def) x.
Proof.
elim=> // hd tl IH def x /= [-> | H]; by [apply my_max_list_prop1 | apply IH].
Qed.

Lemma my_max_list_disj : forall l, disj (t_plus (nat2t 1) (my_max_list l (nat2t O)) :: List.nil) l.
Proof.
move=> l.
apply disj_cons_L; first by apply disj_nil_L.
move/(my_max_list_prop _ (nat2t O)).
move/t_geb_true'.
rewrite -t_gt_t_ge.
move/negP.
apply.
apply t_gtb_true'.
by apply t_gt_1.
Qed.

Lemma t_geb_false : forall a b : t, ~~ t_geb a b -> ~ t_ge a b.
Proof. move=> a b. move/negP => H. by move/t_geb_true'. Qed.

Lemma t_gtb_false : forall a b : t, ~~ t_gtb a b -> ~ t_gt a b.
Proof. move=> a b. move/negP => H. by move/t_gtb_true'. Qed.

Lemma t_geb_false' : forall a b : t, ~ t_ge a b -> ~~ t_geb a b.
Proof. move=> a b H. apply/negP. contradict H. by apply t_geb_true. Qed.

Lemma t_gtb_false' : forall a b : t, ~ t_gt a b -> ~~ t_gtb a b.
Proof. move=> a b H. apply/negP. contradict H. by apply t_gtb_true. Qed.

End IntegralType_Prop.

Inductive negpos : Set :=
| neg : positive -> negpos
| pos : positive -> negpos.

Definition negpos_eqb a b := 
  match (a, b) with  
    | (neg a', neg b') => Ndec.Peqb a' b'
    | (pos a', pos b') => Ndec.Peqb a' b'
    | _ => false
  end.

Lemma negposP : Equality.axiom negpos_eqb.
Proof.
  move=> x y.
  apply: (iffP idP) => H.
  destruct x; destruct y => //.
  rewrite /negpos_eqb /= in H.
  move/positiveP:H => H.
  subst => //.
  rewrite /negpos_eqb /= in H.
  move/positiveP:H => H.
  subst => //.
  subst.
  destruct y; rewrite /negpos_eqb => /=.
  apply/positiveP => //.
  apply/positiveP => //.
Qed.

Canonical Structure negpos_eqMixin := EqMixin negposP.
Canonical Structure negpos_eqType := Eval hnf in EqType _ negpos_eqMixin.

Module ZIT <: IntegralType.

Definition t : eqType := Z_eqType.
Lemma t_dec : forall (x y : t), {x = y} + {x <> y}.
Proof. exact Z_eq_dec. Qed.
Definition nat2t : nat -> t := Z_of_nat.

Definition t_plus : t -> t -> t := Zplus.
Lemma t_plus_assoc : forall a b c, t_plus a (t_plus b c) = t_plus (t_plus a b) c.
Proof. exact Zplus_assoc. Qed.
Lemma t_plus_comm : forall n m, t_plus n m = t_plus m n.
Proof. exact Zplus_comm. Qed.
Lemma t_plus_0 : forall a, t_plus a (nat2t 0) = a.
Proof. move=> a. rewrite /t_plus /nat2t /= Zplus_0_r //. Qed.
Lemma t_plus_1 : forall a, t_plus (nat2t 1) (nat2t a) = nat2t (S a).
Proof. move=> a. rewrite /t_plus /nat2t !Z_S [Z_of_nat 0]/= Zplus_0_l Zplus_comm; done. Qed.

Definition t_minus : t -> t -> t := Zminus.
Definition t_mult : t -> t -> t := Zmult.
Definition t_div : t -> t -> t := Zdiv.
Definition t_mod : t -> t -> t := Zmod.
Definition t_valabs : t -> t := Zabs.

Definition t_ge : t -> t -> Prop := Zge.
Definition t_gt : t -> t -> Prop := Zgt.
Definition t_geb : t -> t -> bool := Zge_bool.
Definition t_gtb : t -> t -> bool := Zgt_bool.

Lemma t_geb_true : forall a b : t, t_geb a b = true -> t_ge a b.
Proof. move=> *; by apply/ZgeP. Qed.
Lemma t_geb_true' : forall a b : t, t_ge a b -> t_geb a b = true.
Proof. move=> *; by apply/ZgeP. Qed.
Lemma t_gtb_true : forall a b : t, t_gtb a b = true -> t_gt a b.
Proof. move=> *; by apply/ZgtP. Qed.
Lemma t_gtb_true' : forall a b : t, t_gt a b -> t_gtb a b = true.
Proof. move=> *; by apply/ZgtP. Qed.

Definition t_eqb : t -> t -> bool := eq_op.
Lemma t_eqb_eq : forall x y : t, t_eqb x y = true -> x = y.
Proof. exact Zeq_bool_eq. Qed.
Lemma t_eqb_refl : forall x : t, t_eqb x x = true.
Proof. move=> *; by apply/ZeqP. Qed.
Lemma t_eqb_false : forall x y : t, t_eqb x y = false <-> x <> y.
Proof. split. by move/eqP. by move/eqP/negbTE. Qed.
Lemma t_eqb_false' : forall x y : t, x <> y -> t_eqb x y = false.
Proof. exact Zeq_bool_neq'. Qed.

Lemma nat2t_inj : forall a b, nat2t a = nat2t b -> a = b.
Proof.
move=> a b.
by move/Z_of_nat_inj.
Qed.

Lemma tP : Equality.axiom t_eqb.
Proof. move=> *; exact: ZeqP. Qed.

Definition t_non_zero : Set := negpos.
Definition t_non_zero2t (x : t_non_zero) : t := 
  match x with neg p => Zneg p | pos p => Zpos p end.
Lemma t_non_zero2t_0 : forall x, t_non_zero2t x <> nat2t 0.
Proof. case=> p //=. Qed.
Definition t2t_non_zero (i : t) (H : i <> nat2t O) : t_non_zero.
Proof.
case:i H => // p H.
exact (pos p).
exact (neg p).
Defined.
Lemma t_non_zero2t_t2t_non_zero : forall z (H : z <> nat2t O), t_non_zero2t (t2t_non_zero z H) = z.
Proof. case=> //. Qed.
Lemma t_non_zero2t_inj : forall x y, t_non_zero2t x = t_non_zero2t y -> x = y.
Proof. case=> p; case=> p' //=; case=> -> //. Qed.

Definition t_non_zero_eqb : t_non_zero -> t_non_zero -> bool := negpos_eqb.
Lemma t_non_zeroP : Equality.axiom t_non_zero_eqb.
Proof. exact negposP. Qed.

Lemma t_ge_refl a : t_ge a a.
Proof. by apply Zle_ge, Zle_refl. Qed.

Lemma t_gt_t_ge a b : (~~ t_gtb a b) = t_geb b a.
Proof. by rewrite /t_gtb Z.gtb_ltb ZltnNge negbK -Z.geb_leb. Qed.

Lemma t_gtb_def a b : t_gtb a b = t_geb a b && (a != b).
Proof.
rewrite /t_gtb /t_geb.
apply/ZgtP.
case: ifP.
  case/andP => /ZgeP X /eqP Y; omega.
move/negP => X Y; apply X.
apply/andP; split.
  apply/ZgeP; omega.
apply/eqP; omega.
Qed.

Lemma t_gt_weak a b : t_gt a b -> t_ge a b.
Proof. rewrite /t_gt /t_ge; omega. Qed.

Lemma t_ge_trans : forall a b c, t_ge a b -> t_ge b c -> t_ge a c.
Proof. exact Zge_trans. Qed.

Lemma t_gt_1 n : t_gt (t_plus (nat2t 1) n) n.
Proof. rewrite /t_gt /t_plus /nat2t [Z_of_nat _]/=; omega. Qed.

(* turn the abstract (A.t_geb _ _) = true, (A.t_ge _ _ ), etc. hypotheses into their implementation,
   so as to be able to use third-party tactics to solve goals (e.g., omega) *)

End ZIT.

Module Import ZIT_prop_m := IntegralType_Prop ZIT.

Ltac open_integral_type_hyp :=
  match goal with
    | |- is_true (ZIT.t_gtb _ _) => apply ZIT.t_gtb_true'; unfold ZIT.t_gt
    | H : is_true (ZIT.t_gtb _ _) |- _ =>
        move: {H}(ZIT.t_gtb_true _ _ H) => H; unfold ZIT.t_gt in H
    | |- is_true (~~ ZIT.t_gtb _ _) => apply t_gtb_false'; unfold ZIT.t_gt
    | H : is_true (~~ ZIT.t_gtb _ _) |- _ =>
        move: {H}(t_gtb_false _ _ H) => H; unfold ZIT.t_gt in H
    | |- is_true (~~ ZIT.t_geb _ _) => apply t_geb_false'; unfold ZIT.t_ge
    | |- is_true (ZIT.t_geb _ _) => apply ZIT.t_geb_true'; unfold ZIT.t_ge
    | H : is_true (ZIT.t_geb _ _) |- _ =>
        move: {H}(ZIT.t_geb_true _ _ H) => H; unfold ZIT.t_ge in H
    | H : is_true (~~ ZIT.t_geb _ _) |- _ =>
        move: {H}(t_geb_false _ _ H) => H; unfold ZIT.t_ge in H
    | |- is_true (ZIT.t_eqb _ _) => apply/ZIT.tP
    | H : is_true (ZIT.t_eqb _ _) |- _ => move/ZIT.tP : H => H

    | H : context [ZIT.t_plus _ _] |- _ => unfold ZIT.t_plus in H
    | H : context [ZIT.t_minus _ _] |- _ => unfold ZIT.t_minus in H
    | H : context [ZIT.t_mult _ _] |- _ => unfold ZIT.t_mult in H
    | H : context [ZIT.t_gt _ _] |- _ => unfold ZIT.t_gt in H
    | H : context [ZIT.t_ge _ _] |- _ => unfold ZIT.t_ge in H
    | H : context [ZIT.nat2t _] |- _ => unfold ZIT.nat2t in H
  end.

Ltac open_integral_type_goal :=
  match goal with
    | |- _ => unfold ZIT.t_plus, ZIT.t_minus, ZIT.t_mult, ZIT.t_gt, ZIT.t_ge, ZIT.nat2t
  end.

Lemma ZIT_ring_theory : @ring_theory ZIT.t Z0 1%Z Zplus Zmult Zminus Zopp (@eq _).
Proof.
constructor.
exact Zplus_0_l.
exact Zplus_comm.
exact Zplus_assoc.
exact Zmult_1_l.
exact Zmult_comm.
exact Zmult_assoc.
exact Zmult_plus_distr_l.
move=> x [] //.
exact Zplus_opp_r.
Qed.

Add Ring ZIT_ring : ZIT_ring_theory.
